"""Injects mkdocs-style `nav` and `page` objects into the HTML jinja2 context.

This generates global and local tables-of-contents (TOCs) usable by (a modified
version of) the mkdocs-material HTML templates.

In particular, for each document, this module generates three separate TOCs:

`global_toc`:
    The `global_toc` is the global table of contents.  If the
    `globaltoc_collapse` theme option is `False`, it contains all documents
    reachable from the root document, as well as any sections of non-root
    documents that contain non-empty TOCs.  If the `globaltoc_collapse` theme
    option is `True`, then the global TOC is restricted to children of:

    - the root,
    - the current document,
    - ancestors of the current document.

`local_toc`:
    The `local_toc` is the local table of contents for the specified document.
    It contains all sections of the current document, but does not contain any
    entries that link outside the current document.

`integrated_local_toc`:
    The `integrated_local_toc` contains all sections of the current document, as
    well as any child documents referenced by a TOC contained in the current
    document.  Whether children of the child docuemnts are included depends on
    the `globaltoc_collapse` theme option.

Background
----------

The Sphinx document model differs from the mkdocs document model in that
documents can be organized as children of other documents and sections within
those documents.

Similar functionality is optionally available in mkdocs-material, through the
`navigation.indexes` feature, which effectively allows documents to be children
of other documents (but not sections within those documents).

However, mkdocs-material specifically documents that `navigation.indexes` is
incompatible with the `toc.integrate` feature.  Furthermore, as noted in
https://github.com/squidfunk/mkdocs-material/issues/3819, the local TOC is
inaccessible in the layered navigation menu used with narrow viewports.  This is
because under the mkdocs-material document model (with `navigation.indexes`
feature enabled), there is no natural way to combine both the local TOC for a
page and the nested list of child documents into a single TOC.

With Sphinx, non-leaf documents are the common case, not a special added
feature, and it is not very reasonable for the TOC to not behave correctly on
such documents.  Furthermore, under the Sphinx document model, child documents
are already organized within the sections of their parent document.  Therefore,
there *is* a natural way to display the local TOC and the nested child documents
as a single TOC --- this combined toc is the `integrated_local_toc`.

The mkdocs-material package uses the global and local TOCs as follows:

Left sidebar:
- Doc 1 (from global_toc)
- Group (from global_toc)
  - Doc 2 (from global_toc)
  - Current page (from global_toc)
    - Section 1 (from local_toc)
    - Section 2 (from local_toc)

Right side bar:
- Section 1 (from local_toc)
- Section 2 (from local_toc)

Note that the local TOC is duplicated into the left sidebar as well, but is
hidden in the normal non-mobile layout, unless the `toc.integrate` feature is
enabled (in that case the right side bar is always hidden).  With a sufficiently
narrow layout, the right side bar is hidden and the duplicate copy of the local
toc in the left sidebar is shown in the layered "hamburger" navigation menu.

The above example is for the case where the current page is a leaf page.  If the
`navigation.indexes` feature is in use and the current page is a non-leaf page,
the sidebars are instead generated as follows:

Left sidebar:
- Doc 1 (from global_toc)
- Group (from global_toc)
  - Doc 2 (from global_toc)
  - Current page (from global_toc)
    - Doc 3 (from global_toc)
    - Doc 4 (from global_toc)

Right side bar:
- Section 1 (from local_toc)
- Section 2 (from local_toc)

In order to support a separate `integrated_local_toc`, this theme modifies the
mkdocs-material templates to generate the sidebars as follows:

Left sidebar:
- Doc 1 (from global_toc)
- Group (from global_toc)
  - Doc 2 (from global_toc)
  - Current page (from global_toc) [class=md-nav__current-nested]
    - Doc 3 (from global_toc)
    - Doc 4 (from global_toc)
  - Current page (from global_toc) [class=md-nav__current-toc]
    - Section 1 (from local_toc_integrated)
    - Section 2 (from local_toc_integrated)

Right side bar:
- Section 1 (from local_toc)
- Section 2 (from local_toc)

The left sidebar contains two copies of the local toc, one generated from
`global_toc` and the other from `local_toc_integrated`, but CSS rules based on
the added `md-nav__current-nested` and `md-nav__current-toc` ensure that at most
one copy is shown at a time.

The root document is an exception: in Sphinx the global document structure is
defined by adding `toctree` nodes to the root document.  Technically those
`toctree` nodes are still contained within the usual section structure of the
root document, but the built-in TOC functionality in Sphinx treats the root
document specially, and extacts any `toctree` nodes into a separate global TOC
hierarchy, independent of the section structure of the root document.  In
practice, users often place the `toctree` nodes at the end of the root document,
effectively making them children of the last section, but it is not intended
that they are actually a part of any section.  Therefore, for the root document
there is no natural way to integrate the local and global TOCs, and consequently
the local TOC is simply unavailable when the `toc.integrate` feature is enabled
or when using the "layered" navigation menu.

"""

import collections
import copy
import os
import re
from typing import (
    cast,
    List,
    Union,
    NamedTuple,
    Optional,
    Tuple,
    Iterator,
    Dict,
    Iterable,
    Set,
)
import urllib.parse
import docutils.nodes
import markupsafe
from sphinx.addnodes import docutils_meta, meta as sphinx_meta
import sphinx.builders
import sphinx.builders.html
import sphinx.application
import sphinx.environment.adapters.toctree
import sphinx.util.docutils
import sphinx.util.osutil

from .apidoc import object_description_options

StandaloneHTMLBuilder = sphinx.builders.html.StandaloneHTMLBuilder


# env var is only defined in RTD hosted builds
READTHEDOCS = os.getenv("READTHEDOCS")


def _strip_fragment(url: str) -> str:
    """Returns the url with any fragment identifier removed."""
    return re.sub("#.*", "", url)


def _insert_wbr(text: str) -> str:
    """Inserts <wbr> tags after likely split points for API symbols."""
    # Split after punctuation
    text = re.sub("([.:_-]+)", r"\1<wbr>", text)
    # Split before brackets
    text = re.sub(r"([(\[{])", r"<wbr>\1", text)
    # Split between camel-case words
    text = re.sub(r"([a-z])([A-Z])", r"\1<wbr>\2", text)
    return text


class MkdocsNavEntry:
    # Title to display, as HTML.
    title: str

    # Aria label text, plain text.
    aria_label: Optional[str] = None

    # URL of this page, or the first descendent if `caption_only` is `True`.
    url: Optional[str]
    # List of children
    children: List["MkdocsNavEntry"]
    # Set to `True` if this page, or a descendent, is the current page.
    # Excludes links to sections within in an active page.
    active: bool
    # Set to `True` if this page is the current page.  Excludes links to
    # sections within an active page.
    current: bool

    # Set to `True` if `active`, or if this is a link to a section within an `active` page.
    active_or_section_within_active: bool = False

    # Set to `True` if this entry does not refer to a unique page but is merely
    # a TOC caption.
    caption_only: bool

    def __init__(self, title_text: str, **kwargs):
        self.__dict__.update(kwargs)
        self.title = f'<span class="md-ellipsis">{_insert_wbr(title_text)}</span>'
        if not self.aria_label:
            self.aria_label = title_text

    def __repr__(self):
        return repr(self.__dict__)


class _TocVisitor(docutils.nodes.NodeVisitor):
    """NodeVisitor used by `_get_mkdocs_toc`."""

    def __init__(
        self,
        document: docutils.nodes.document,
        builder: sphinx.builders.html.StandaloneHTMLBuilder,
    ):
        super().__init__(document)
        self._prev_caption: Optional[docutils.nodes.Element] = None
        self._rendered_title_text: Optional[str] = None
        self._url: Optional[str] = None
        self._builder = builder
        # List of direct children.
        self._children: List[MkdocsNavEntry] = []

    def _render(self, node: Union[docutils.nodes.Node, List[docutils.nodes.Node]]):
        """Returns the HTML representation of `node`."""
        if not isinstance(node, list):
            node = [node]
        return "".join(self._builder.render_partial(x)["fragment"] for x in node)

    def _render_title(
        self, node: Union[docutils.nodes.Node, List[docutils.nodes.Node]]
    ):
        """Returns the text representation of `node`."""
        if not isinstance(node, list):
            node = [node]
        return str(markupsafe.Markup.escape("".join(x.astext() for x in node)))

    def visit_reference(self, node: docutils.nodes.reference):
        self._rendered_title_text = self._render_title(node.children)
        self._url = node.get("refuri")
        raise docutils.nodes.SkipChildren

    # `only` directives can result in `comment` nodes.
    def visit_comment(self, node: docutils.nodes.comment):
        raise docutils.nodes.SkipChildren

    def visit_compact_paragraph(self, node: docutils.nodes.Element):
        pass

    def visit_toctree(self, node: docutils.nodes.Node):
        raise docutils.nodes.SkipChildren

    def visit_paragraph(self, node: docutils.nodes.Node):
        pass

    # In sphinx < 3.5.4, TOC captions are represented using a caption node.
    def visit_caption(self, node: docutils.nodes.caption):
        self._prev_caption = node
        raise docutils.nodes.SkipChildren

    # In sphinx >= 3.5.4, TOC captions are represented using a title node.
    def visit_title(self, node: docutils.nodes.title):
        self._prev_caption = node
        raise docutils.nodes.SkipChildren

    def visit_bullet_list(self, node: docutils.nodes.bullet_list):
        if self._prev_caption is not None and self._prev_caption.parent is node.parent:
            # Insert as sub-entry of the previous caption.
            title_text = self._render_title(self._prev_caption.children)
            self._prev_caption = None
            child_visitor = _TocVisitor(self.document, self._builder)
            node.walk(child_visitor)
            url = None
            children = child_visitor._children
            if children:
                url = children[0].url
            self._children.append(
                MkdocsNavEntry(
                    title_text=title_text,
                    url=url,
                    children=children,
                    active=False,
                    current=False,
                    caption_only=True,
                )
            )
            raise docutils.nodes.SkipChildren
        # Otherwise, just process each list_item as direct children.

    def get_result(self) -> MkdocsNavEntry:
        return MkdocsNavEntry(
            title_text=cast(str, self._rendered_title_text),
            url=self._url,
            children=self._children,
            active=False,
            current=False,
            caption_only=False,
        )

    def visit_list_item(self, node: docutils.nodes.list_item):
        # Child node.  Collect its url, title, and any children using a separate
        # `_TocVisitor`.
        child_visitor = _TocVisitor(self.document, self._builder)
        for child in node.children:
            child.walk(child_visitor)
        child_result = child_visitor.get_result()
        self._children.append(child_result)
        raise docutils.nodes.SkipChildren


def _get_mkdocs_toc(
    toc_node: Optional[docutils.nodes.Node],
    builder: sphinx.builders.html.StandaloneHTMLBuilder,
) -> List[MkdocsNavEntry]:
    """Converts a docutils toc node into a mkdocs-format JSON toc."""
    visitor = _TocVisitor(sphinx.util.docutils.new_document(""), builder)

    # toc_node can be None for projects with no toctree or 1 rst-file only.
    if toc_node is not None:
        toc_node.walk(visitor)
    return visitor._children


class _NavContextObject(list):
    homepage: dict


def _traverse_mkdocs_toc(toc: List[MkdocsNavEntry]) -> Iterator[MkdocsNavEntry]:
    for entry in toc:
        yield entry
        yield from _traverse_mkdocs_toc(entry.children)


def _relative_uri_to_root_relative_and_anchor(
    builder: sphinx.builders.html.StandaloneHTMLBuilder,
    base_pagename: str,
    relative_uri: str,
) -> Optional[Tuple[str, str]]:
    """Converts a relative URI to a root-relative uri and anchor."""
    uri = urllib.parse.urlparse(
        urllib.parse.urljoin(builder.get_target_uri(base_pagename), relative_uri)
    )
    if uri.netloc:
        return None
    return (uri.path, uri.fragment)


class DomainAnchorEntry(NamedTuple):
    domain_name: str
    name: str
    display_name: str
    objtype: str
    priority: int
    synopsis: Optional[str]


def _make_domain_anchor_map(
    env: sphinx.environment.BuildEnvironment,
) -> Dict[Tuple[str, str], DomainAnchorEntry]:
    builder = cast(sphinx.builders.Builder, env.app.builder)
    docname_to_url = {
        docname: builder.get_target_uri(docname) for docname in env.found_docs
    }
    m: Dict[Tuple[str, str], DomainAnchorEntry] = {}
    for domain_name, domain in env.domains.items():
        synopses: Dict[Tuple[str, str], str] = {}
        get_object_synopses = getattr(domain, "get_object_synopses", None)
        if get_object_synopses is not None:
            for key, synopsis in get_object_synopses():
                synopses.setdefault(key, synopsis)
        for (
            name,
            dispname,
            objtype,
            docname,
            anchor,
            priority,
        ) in domain.get_objects():
            if domain_name == "std" and objtype == "doc":
                # Don't add an extra tooltip for plain documents.
                continue
            url = docname_to_url.get(docname)
            if url is None:
                continue
            key = (url, anchor)
            synopsis = synopses.get((docname, anchor))
            m.setdefault(
                key,
                DomainAnchorEntry(
                    domain_name, name, dispname, objtype, priority, synopsis
                ),
            )
    return m


def _get_domain_anchor_map(
    app: sphinx.application.Sphinx,
) -> Dict[Tuple[str, str], DomainAnchorEntry]:
    key = "sphinx_immaterial_domain_anchor_map"
    env = app.env
    assert env is not None
    m = getattr(env, key, None)
    if m is None:
        m = _make_domain_anchor_map(env)
        setattr(env, key, m)
    return m


def _add_domain_info_to_toc(
    app: sphinx.application.Sphinx, toc: List[MkdocsNavEntry], pagename: str
) -> None:
    m = _get_domain_anchor_map(app)
    assert isinstance(app.builder, StandaloneHTMLBuilder)
    env = app.env
    assert env is not None
    for entry in _traverse_mkdocs_toc(toc):
        if entry.caption_only or entry.url is None:
            continue
        refinfo = _relative_uri_to_root_relative_and_anchor(
            app.builder, pagename, entry.url
        )
        if refinfo is None:
            continue
        objinfo = m.get(refinfo)
        if objinfo is None:
            continue
        domain = env.domains[objinfo.domain_name]
        label = domain.get_type_name(domain.object_types[objinfo.objtype])
        options = object_description_options.get_object_description_options(
            env, objinfo.domain_name, objinfo.objtype
        )
        tooltip = object_description_options.format_object_description_tooltip(
            env, options, objinfo.name, objinfo.synopsis
        )
        toc_icon_text = options["toc_icon_text"]
        toc_icon_class = options["toc_icon_class"]
        title_prefix = ""
        if toc_icon_text is not None and toc_icon_class is not None:
            title_prefix = (
                f'<span aria-label="{label}" '
                f'class="objinfo-icon objinfo-icon__{toc_icon_class}" '
                f'title="{label}">{toc_icon_text}</span>'
            )
        span_prefix = "<span "
        assert entry.title.startswith(span_prefix)
        entry.title = (
            title_prefix
            + f'<span title="{markupsafe.Markup.escape(tooltip)}" '
            + entry.title[len(span_prefix) :]
        )


def _get_current_page_in_toc(toc: List[MkdocsNavEntry]) -> Optional[MkdocsNavEntry]:
    for entry in toc:
        if not entry.active:
            continue
        if entry.current:
            return entry
        return _get_current_page_in_toc(entry.children)
    return None


def _prune_toc_by_active(
    entry: MkdocsNavEntry, active: bool
) -> Optional[MkdocsNavEntry]:
    """Prunes entries from the TOC tree according to whether they are active.

    Any TOC entries with a target on the current page (i.e. a section within the
    current page) are marked active, while entries with a target on a different
    page are not marked active.

    :param entry: TOC root to recursively prune.
    :param active: If `True`, prune targets not on the current page.  If
      `False`, prune targets on the current page, except if they transitively
      contain children not in the current page.
    :returns: Pruned copy of `entry`.
    """
    if active and not entry.active_or_section_within_active:
        return None

    entry = copy.copy(entry)

    new_children = []
    for child in entry.children:
        new_child = _prune_toc_by_active(child, active)
        if new_child is not None:
            new_children.append(new_child)
    entry.children = new_children

    if entry.active_or_section_within_active and not active and not entry.children:
        return None

    return entry


TocEntryKey = Tuple[int, ...]


def _build_toc_index(
    toc: List[MkdocsNavEntry],
) -> Tuple[Dict[str, List[TocEntryKey]], Set[TocEntryKey]]:
    """Builds a map from URL to list of toc entries.

    This is used by `_get_global_toc` to efficiently prune the cached TOC for a
    given page.
    """
    url_map: Dict[str, List[TocEntryKey]] = collections.defaultdict(list)

    global_toc_keys: Set[TocEntryKey] = set()

    def _traverse(
        entries: List[MkdocsNavEntry],
        parent_key: TocEntryKey,
        parent_url: Optional[str],
    ):
        for i, entry in enumerate(entries):
            child_key = parent_key + (i,)
            url: Optional[str] = None
            if not entry.caption_only:
                url = entry.url
                if url is not None:
                    url = _strip_fragment(url)
                    url_map[url].append(child_key)
            _traverse(entry.children, child_key, url)
            if url != parent_url or child_key in global_toc_keys:
                global_toc_keys.add(child_key)
                global_toc_keys.add(parent_key)

    _traverse(toc, (), None)
    return url_map, global_toc_keys


_FAKE_DOCNAME = "fakedoc"


class CachedTocInfo:
    """Cached representation of the global TOC.

    This is generated once (per writer process) and re-used for all pages.

    Obtaining the global TOC via `TocTree.get_toctree_for` is expensive because
    we first have to obtain a complete TOC of all pages, and then prune it for
    the current page.  The overall cost to generate the TOCs for all pages
    therefore ends up being quadratic in the number of pages, and in practice as
    the number of pages reaches several hundred, a significant fraction of the
    total documentation generation time is due to the TOC.

    By the caching the TOC and a `url_map` that allows to efficiently prune the
    TOC for a given page, the cost of generating the TOC for each page is much
    lower.
    """

    def __init__(self, app: sphinx.application.Sphinx):
        # Sphinx always generates a TOC relative to a particular page, and
        # converts all page references to relative URLs.  Use an empty string as
        # the page name to ensure the relative URLs that Sphinx generates are
        # relative to the base URL.  When generating the per-page TOCs from this
        # cached data structure the URLs will be converted to be relative to the
        # current page.
        env = app.env
        assert env is not None
        builder = app.builder
        assert isinstance(builder, StandaloneHTMLBuilder)
        global_toc_node = sphinx.environment.adapters.toctree.TocTree(
            env
        ).get_toctree_for(
            _FAKE_DOCNAME,
            builder,
            collapse=False,
            maxdepth=-1,
            titles_only=False,
        )
        global_toc = _get_mkdocs_toc(global_toc_node, builder)
        _add_domain_info_to_toc(app, global_toc, _FAKE_DOCNAME)
        self.entries = global_toc
        self.url_map, self.global_toc_keys = _build_toc_index(global_toc)


def _get_cached_globaltoc_info(app: sphinx.application.Sphinx) -> CachedTocInfo:
    """Obtains the cached global TOC, generating it if necessary."""
    key = "sphinx_immaterial_global_toc_cache"
    data = getattr(app.env, key, None)
    if data is not None:
        return data
    data = CachedTocInfo(app)
    setattr(app.env, key, data)
    return data


def _get_ancestor_keys(keys: Iterable[TocEntryKey]) -> Set[TocEntryKey]:
    ancestors = set()
    for key in keys:
        while key not in ancestors:
            ancestors.add(key)
            key = key[:-1]
    return ancestors


def _get_global_toc(app: sphinx.application.Sphinx, pagename: str, collapse: bool):
    """Obtains the global TOC for a given page."""
    cached_data = _get_cached_globaltoc_info(app)
    builder = app.builder
    assert isinstance(builder, StandaloneHTMLBuilder)
    fake_page_url = builder.get_target_uri(_FAKE_DOCNAME)
    fake_relative_url = sphinx.util.osutil.relative_uri(
        fake_page_url, builder.get_target_uri(pagename)
    )
    keys = set(cached_data.url_map[fake_relative_url])
    global_toc_keys = cached_data.global_toc_keys
    ancestors = _get_ancestor_keys(keys)
    real_page_url = builder.get_target_uri(pagename)

    def _make_toc_for_page(key: TocEntryKey, children: List[MkdocsNavEntry]):
        page_is_current = key in keys
        new_children: List[MkdocsNavEntry] = []
        for i, child in enumerate(children):
            child_key = key + (i,)
            in_ancestors = child_key in ancestors
            child_active = in_ancestors
            child_current = in_ancestors and child_key in keys
            if (
                not child_active
                and not page_is_current
                and child_key not in global_toc_keys
            ):
                continue
            child = copy.copy(child)
            new_children.append(child)
            if child.url is not None:
                root_relative_url = urllib.parse.urljoin(fake_page_url, child.url)
                uri = urllib.parse.urlparse(root_relative_url)
                if not uri.netloc:
                    child.url = sphinx.util.osutil.relative_uri(
                        real_page_url, root_relative_url
                    )
                    if uri.fragment or child.url == "":
                        child.url += f"#{uri.fragment}"
            child.active = child_active and not page_is_current
            child.current = child_current and not page_is_current
            child.active_or_section_within_active = child_active
            if in_ancestors or child.caption_only or not collapse:
                child.children = _make_toc_for_page(child_key, child.children)
            else:
                child.children = []
        return new_children

    return _make_toc_for_page((), cached_data.entries)


def _get_mkdocs_tocs(
    app: sphinx.application.Sphinx,
    pagename: str,
    duplicate_local_toc: bool,
    toc_integrate: bool,
) -> Tuple[List[MkdocsNavEntry], List[MkdocsNavEntry], List[MkdocsNavEntry]]:
    """Generates the global and local TOCs for a document.

    :param app: The sphinx application object.
    :param pagename: The name of the document for which to generate the tocs.
    :param duplicate_local_toc: Duplicate the local toc in the global toc.
    :param toc_integrate: Indicates if the `toc.integrate` feature is enabled.
    :returns: A tuple `(global_toc, local_toc, integrated_local_toc)`.
    """
    theme_options = app.config["html_theme_options"]
    global_toc = _get_global_toc(
        app=app,
        pagename=pagename,
        collapse=theme_options.get("globaltoc_collapse", False),
    )
    local_toc: List[MkdocsNavEntry] = []
    integrated_local_toc: List[MkdocsNavEntry] = []
    env = app.env
    assert env is not None
    builder = app.builder
    assert isinstance(builder, StandaloneHTMLBuilder)
    if pagename != env.config.master_doc:
        # Extract entry from `global_toc` corresponding to the current page.
        current_page_toc_entry = _get_current_page_in_toc(global_toc)
        if current_page_toc_entry is not None:
            integrated_local_toc = [copy.copy(current_page_toc_entry)]
            integrated_local_toc[0].children = list(integrated_local_toc[0].children)
            if not toc_integrate:
                local_toc = cast(
                    List[MkdocsNavEntry],
                    [_prune_toc_by_active(current_page_toc_entry, active=True)],
                )
            if toc_integrate:
                current_page_toc_entry.children = []
            elif not duplicate_local_toc:
                current_page_toc_entry.children = [
                    child
                    for child in [
                        _prune_toc_by_active(child, active=False)
                        for child in current_page_toc_entry.children
                    ]
                    if child is not None
                ]
    else:
        # Every page is a child of the root page.  We still want a full TOC
        # tree, though.
        local_toc_node = sphinx.environment.adapters.toctree.TocTree(env).get_toc_for(
            pagename,
            builder,
        )
        local_toc = _get_mkdocs_toc(local_toc_node, builder)
        _add_domain_info_to_toc(app, local_toc, pagename)

    if len(local_toc) == 1 and len(local_toc[0].children) == 0:
        local_toc = []

    if len(integrated_local_toc) == 1 and len(integrated_local_toc[0].children) == 0:
        integrated_local_toc = []

    return global_toc, local_toc, integrated_local_toc


def _html_page_context(
    app: sphinx.application.Sphinx,
    pagename: str,
    templatename: str,
    context: dict,
    doctree: docutils.nodes.Node,
) -> None:
    env = app.env
    assert env is not None

    theme_options: dict = app.config["html_theme_options"]
    features = theme_options.get("features", ())
    assert isinstance(features, collections.abc.Sequence)
    page_title = markupsafe.Markup.escape(
        markupsafe.Markup(context.get("title")).striptags()
    )
    meta = context.get("meta")
    if meta is None:
        meta = {}
    global_toc, local_toc, integrated_local_toc = _get_mkdocs_tocs(
        app,
        pagename,
        duplicate_local_toc=isinstance(meta.get("duplicate-local-toc"), str),
        toc_integrate="toc.integrate" in features,
    )
    context.update(nav=_NavContextObject(global_toc))
    context["nav"].homepage = dict(
        url=context["pathto"](context["master_doc"]),
    )

    toc_title = theme_options.get("toc_title")
    if toc_title:
        toc_title = str(markupsafe.Markup.escape(toc_title))
    elif (
        theme_options.get("toc_title_is_page_title")
        and local_toc
        and len(local_toc) == 1
    ):
        # Use single top-level heading as table of contents heading.
        toc_title = local_toc[0].title

    context.update(
        config={
            "mdx_configs": {
                "toc": {"title": toc_title},
            },
            "extra": {"consent": theme_options.get("consent")},
        }
    )

    if len(local_toc) == 1:
        # If there is a single top-level heading, it is treated as the page
        # heading, and it would be redundant to also include it as an entry in
        # the local toc.
        local_toc = local_toc[0].children

    if len(integrated_local_toc) == 1:
        # If there is a single top-level heading, it is treated as the page
        # heading, and it would be redundant to also include it as an entry in
        # the local toc.
        integrated_local_toc = integrated_local_toc[0].children

    # Add other context values in mkdocs/mkdocs-material format.
    page = dict(
        title=page_title,
        is_homepage=(pagename == context["master_doc"]),
        toc=local_toc,
        integrated_local_toc=integrated_local_toc,
        meta={"hide": [], "revision_date": context.get("last_updated"), "meta": []},
        content=context.get("body"),
    )
    if doctree is not None:
        # extract meta nodes from document node only (not descendants)
        meta_tags = [
            doc_node
            for doc_node in doctree.document.children
            if isinstance(doc_node, (docutils_meta, sphinx_meta))
        ]
        for tag in meta_tags:
            # feed them into the mkdocs template context
            attrs = {key: value for key, value in tag.attributes.items() if value}
            page["meta"]["meta"].append(attrs)
    if meta.get("tocdepth") == 0 or "hide-toc" in meta:
        page["meta"]["hide"].append("toc")
    if "hide-navigation" in meta:
        page["meta"]["hide"].append("navigation")
    if "hide-footer" in meta:
        page["meta"]["hide"].append("footer")
    if "hide-feedback" in meta:
        page["meta"]["hide"].append("feedback")
    if "show-comments" in meta:
        page["meta"]["comments"] = True
    if context.get("next"):
        page["next_page"] = {
            "title": markupsafe.Markup.escape(
                markupsafe.Markup(context["next"]["title"]).striptags()
            ),
            "url": context["next"]["link"],
        }
    if context.get("prev"):
        page["previous_page"] = {
            "title": markupsafe.Markup.escape(
                markupsafe.Markup(context["prev"]["title"]).striptags()
            ),
            "url": context["prev"]["link"],
        }
    repo_url: Optional[str] = theme_options.get("repo_url")
    edit_uri: Optional[str] = theme_options.get("edit_uri")
    if repo_url and edit_uri and not READTHEDOCS and "hide-edit-link" not in meta:
        page["edit_url"] = "/".join(
            [
                repo_url.rstrip("/"),
                edit_uri.strip("/"),
                env.doc2path(pagename, False),
            ]
        )
    context.update(
        page=page,
    )


def setup(app: sphinx.application.Sphinx):
    app.connect("html-page-context", _html_page_context)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import pathlib

import pytest

import docutils.nodes
import sphinx.domains.python

from sphinx.testing.path import path as SphinxPath

pytest_plugins = ("sphinx.testing.fixtures",)


@pytest.fixture
def theme_make_app(tmp_path: pathlib.Path, make_app):

    conf = """
extensions = [
    "sphinx_immaterial",
]
html_theme = "sphinx_immaterial"
"""

    def make(extra_conf: str = "", **kwargs):
        (tmp_path / "conf.py").write_text(conf + extra_conf, encoding="utf-8")
        (tmp_path / "index.rst").write_text("", encoding="utf-8")
        return make_app(srcdir=SphinxPath(str(tmp_path)), **kwargs)

    yield make


def test_transform_type_annotations_pep604(theme_make_app):
    app = theme_make_app(
        confoverrides=dict(),
    )

    for annotation, expected_text in [
        ("Union[int, float]", "int | float"),
        ("Literal[1, 2, None]", "1 | 2 | None"),
    ]:
        parent = docutils.nodes.TextElement("", "")
        parent.extend(sphinx.domains.python._parse_annotation(annotation, app.env))
        assert parent.astext() == expected_text



"""Adds support for obtaining property types from docstring signatures, and
improves formatting by PyProperty of type annotations."""
import re
from typing import Tuple, Optional, Any

import sphinx.addnodes
import sphinx.domains
import sphinx.domains.python
import sphinx.ext.autodoc
import sphinx.util.inspect
import sphinx.util.typing

PropertyDocumenter = sphinx.ext.autodoc.PropertyDocumenter


property_sig_re = re.compile("^(\\(.*)\\)\\s*->\\s*(.*)$")


def _get_property_getter(obj: Any) -> Any:
    # property
    func = sphinx.util.inspect.safe_getattr(obj, "fget", None)
    if func is None:
        # cached_property
        func = sphinx.util.inspect.safe_getattr(obj, "func", None)
    return func


def _get_return_type_from_fget_doc(obj: Any) -> Optional[str]:
    fget = _get_property_getter(obj)
    if fget is None:
        return None
    doc = sphinx.util.inspect.safe_getattr(fget, "__doc__", None)
    if doc is None:
        return None
    line = doc.splitlines()[0]
    line = line.rstrip("\\").strip()
    match = property_sig_re.match(line)
    if not match:
        return None
    _, retann = match.groups()
    return retann


def apply_property_documenter_type_annotation_fix():

    # Modify PropertyDocumenter to support obtaining signature from docstring.

    orig_import_object = PropertyDocumenter.import_object

    def import_object(self: PropertyDocumenter, raiseerror: bool = False) -> bool:
        result = orig_import_object(self, raiseerror)
        if not result:
            return False

        func = _get_property_getter(self.object)

        if func is not None:
            try:
                signature = sphinx.util.inspect.signature(
                    func, type_aliases=self.config.autodoc_type_aliases
                )
                if (
                    signature.return_annotation
                    is not sphinx.util.inspect.Parameter.empty
                ):
                    self.retann = sphinx.util.typing.stringify(
                        signature.return_annotation
                    )
                return True
            except:  # pylint: disable=bare-except
                pass

        if not self.retann:
            new_retann = _get_return_type_from_fget_doc(self.object)
            if new_retann is not None:
                self.retann = new_retann
        return True

    PropertyDocumenter.import_object = import_object

    old_add_directive_header = PropertyDocumenter.add_directive_header

    def add_directive_header(self, sig: str) -> None:
        start_line = len(self.directive.result.data)
        old_add_directive_header(self, sig)

        # Check for return annotation
        retann = self.retann or _get_return_type_from_fget_doc(self.object)
        if retann is None:
            return

        # Check if type annotation has already been added.
        type_line_prefix = self.indent + "   :type: "
        for line in self.directive.result.data[start_line:]:
            if line.startswith(type_line_prefix):
                return

        # Type annotation not already added.
        self.add_line("   :type: " + retann, self.get_sourcename())

    PropertyDocumenter.add_directive_header = add_directive_header

    # Modify PyProperty to improve formatting of :type: option
    PyProperty = sphinx.domains.python.PyProperty

    def handle_signature(
        self, sig: str, signode: sphinx.addnodes.desc_signature
    ) -> Tuple[str, str]:
        fullname, prefix = super(PyProperty, self).handle_signature(sig, signode)

        typ = self.options.get("type")
        if typ:
            signode += sphinx.addnodes.desc_sig_punctuation("", " : ")
            signode += sphinx.domains.python._parse_annotation(typ, self.env)

        return fullname, prefix

    PyProperty.handle_signature = handle_signature


apply_property_documenter_type_annotation_fix()



"""Extension that adds `highlight-push` and `highlight-pop` directives."""

from typing import List

import docutils.nodes
import sphinx.addnodes
import sphinx.application
import sphinx.util.docutils
import sphinx.util.logging
from sphinx.transforms.post_transforms.code import HighlightLanguageVisitor

logger = sphinx.util.logging.getLogger(__name__)


class HighlightPushDirective(sphinx.util.docutils.SphinxDirective):

    has_content = False
    required_arguments = 0
    optional_arguments = 0

    def run(self) -> List[docutils.nodes.Node]:

        stack = self.env.temp_data.setdefault("highlight_language_stack", [])
        stack.append(self.env.temp_data.get("highlight_language"))
        return [sphinx.addnodes.highlightlang(push=True)]


class HighlightPopDirective(sphinx.util.docutils.SphinxDirective):

    has_content = False
    required_arguments = 0
    optional_arguments = 0

    def run(self) -> List[docutils.nodes.Node]:

        stack = self.env.temp_data.get("highlight_language_stack")
        if not stack:
            logger.error(
                "`highlight-pop` directive not preceded by `highlight-push` directive",
                location=self.get_location(),
            )
            return []
        language = stack.pop()
        if language is None:
            self.env.temp_data.pop("highlight_language", None)
        else:
            self.env.temp_data["highlight_language"] = language
        return [sphinx.addnodes.highlightlang(pop=True)]


def _monkey_patch_highlight_language_visitor():

    orig_visit_highlightlang = HighlightLanguageVisitor.visit_highlightlang

    def visit_highlightlang(self, node: sphinx.addnodes.highlightlang) -> None:
        if "push" in node.attributes:
            self.settings.append(self.settings[-1])
        elif "pop" in node.attributes:
            if self.settings:
                self.settings.pop()
            else:
                logger.error(
                    "highlight-pop node not preceded by hihglight-push node",
                    location=node,
                )
            if not self.settings:
                self.settings.append(self.default_setting)
                logger.error(
                    "highlight-pop node not preceded by hihglight-push node",
                    location=node,
                )
        else:
            orig_visit_highlightlang(self, node)

    HighlightLanguageVisitor.visit_highlightlang = visit_highlightlang


_monkey_patch_highlight_language_visitor()


def setup(app: sphinx.application.Sphinx):
    app.add_directive("highlight-push", HighlightPushDirective)
    app.add_directive("highlight-pop", HighlightPopDirective)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""This module inherits from the generic ``admonition`` directive and makes the
title optional."""
from abc import ABC
import io
import hashlib
from pathlib import Path, PurePath
import re
from typing import List, Dict, Any, Tuple, Optional, Type, Union
from docutils import nodes
from docutils.parsers.rst import directives, Directive
import jinja2
import pydantic
import sphinx.addnodes
from sphinx.application import Sphinx
from sphinx.config import Config
from sphinx.environment import BuildEnvironment
import sphinx.ext.todo
from sphinx.locale import admonitionlabels, _
from sphinx.util.logging import getLogger
from sphinx.writers.html5 import HTML5Translator
from .inline_icons import load_svg_into_builder_env

logger = getLogger(__name__)

# treat the todo directive from the sphinx extension as a built-in directive
admonitionlabels["todo"] = _("Todo")


class CustomAdmonitionConfig(pydantic.BaseModel):
    """This data class validates the user's configuration value(s) in
    :confval:`sphinx_immaterial_custom_admonitions`. Each validated object
    corresponds to a generated custom :rst:dir:`admonition` directive tailored with
    theme specific options.
    """

    name: str
    """The required name of the directive. This will be also used as a CSS class name.
    This value shall have characters that match the regular expression pattern
    ``[a-zA-Z0-9_-]``."""
    title: Optional[str]
    """The default title to use when rendering the custom admonition. If this is
    not specified, then the `name` value is converted and used."""
    icon: Optional[str] = None
    """The relative path to an icon that will be used in the admonition's
    title. If specified, this path shall be relative to

    - a SVG file placed in the documentation project's
      :confval:`sphinx_immaterial_icon_path` (this takes precedence).
    - the ``.icons`` folder that has `all of the icons bundled with this theme
      <https://github.com/squidfunk/mkdocs-material/tree/master/material/.icons>`_."""
    color: Optional[Tuple[int, int, int]] = None
    """The base color to be used for the admonition's styling. If specified, this
    must be specified as a RGB color space. Each color component shall be
    confined to the range [0, 255]. It is also possible to use the
    :py:mod:`colorsys` module if converting from a different color space (like
    HSL) to RGB."""
    classes: Optional[List[str]]
    """If specified, this list of qualified names will be added to every rendered
    admonition (specific to the generated directive) element's ``class`` attribute.

    To adopt the styling of pre-existing admonition, include the desired
    admonition directive's name in this list.

    .. code-block:: python
        :caption: Adopting the style of an :dutree:`hint` admonition:

        sphinx_immaterial_custom_admonitions = [
            {
                "name": "my-admonition",
                "classes": ["hint"],
            },
        ]
    """
    override: bool = False
    """Can be used to override an existing directive (default is :python:`False`). Only
    set this to :python:`True` if the directive being overridden is an existing
    admonition :ref:`defined by rST and Sphinx <predefined_admonitions>` or an
    admonition :ref:`inherited from the mkdocs-material theme <inherited_admonitions>`.
    """

    # pylint: disable=no-self-argument

    @pydantic.validator("name", pre=True)
    def validate_name(cls, val):
        illegal = re.findall(r"([^a-zA-Z0-9\-_])", val)
        if illegal:
            raise pydantic.ValidationError(
                f"The following characters are illegal for directive names: {illegal}"
            )
        return val

    @pydantic.validator("title", always=True)
    def validate_title(cls, val, values):
        if val is None:
            val = " ".join(re.split(r"[\-_]+", values["name"])).title()
        return val

    @pydantic.validator("color", each_item=True)
    def validate_color_component(cls, val):
        if 0 <= val <= 255:
            return val
        raise pydantic.ValidationError(
            f"color component {val} is not in range [0, 255]"
        )

    @pydantic.validator("classes", each_item=True)
    def conform_classes(cls, val):
        return nodes.make_id(val)

    # pylint: enable=no-self-argument


def patch_visit_admonition():
    orig_func = HTML5Translator.visit_admonition

    def visit_admonition(self: HTML5Translator, node: nodes.Element, name: str = ""):
        collapsible: Optional[str] = node.get("collapsible", None)
        if collapsible is not None:
            tag_extra_args: Dict[str, Any] = dict(CLASS="admonition")
            if collapsible.lower() == "open":
                tag_extra_args["open"] = ""
            self.body.append(self.starttag(node, "details", **tag_extra_args))
            title = node.children[0]
            assert isinstance(title, nodes.title)
            self.body.append(
                self.starttag(title, "summary", suffix="", CLASS="admonition-title")
            )
            for child in title.children:
                child.walkabout(self)
            self.body.append("</summary>")
            del node.children[0]
        else:
            orig_func(self, node, name)

    HTML5Translator.visit_admonition = visit_admonition


def patch_depart_admonition():
    orig_func = HTML5Translator.depart_admonition

    def depart_admonition(self: HTML5Translator, node: Optional[nodes.Element] = None):
        if node is None or node.get("collapsible", None) is None:
            orig_func(self, node)
        else:
            self.body.append("</details>\n")

    HTML5Translator.depart_admonition = depart_admonition


class CustomAdmonitionDirective(Directive, ABC):
    """A base class to define custom admonition directives.

    .. warning::
        Do not instantiate an object directly from this class as it could mess up the
        argument parsing for other derivative classes.
    """

    node_class: Type[nodes.Admonition] = nodes.admonition
    optional_arguments: int
    final_argument_whitespace: bool = True
    has_content = True
    default_title: str = ""
    classes: List[str] = []
    option_spec = {
        "class": directives.class_option,
        "name": directives.unchanged,
        "collapsible": directives.unchanged,
        "no-title": directives.flag,
        "title": directives.unchanged,
    }

    def run(self):
        # docutils.parsers.rst.roles.set_classes() is deprecated &
        # its replacement is not available in older versions, so
        # manually convert key from "class" to "classes"
        if "class" in self.options:
            if "classes" not in self.options:
                self.options["classes"] = []
            self.options["classes"].extend(self.options["class"])
            del self.options["class"]
        title_text = "" if not self.arguments else self.arguments[0]
        if "title" in self.options:
            # this option can be combined with the directive argument used as a title.
            title_text += (" " if title_text else "") + self.options["title"]
            # don't auto-assert `:no-title:` if value is blank; just use default
        if not title_text:
            title_text = self.default_title
        self.assert_has_content()
        admonition_node = self.node_class("\n".join(self.content), **self.options)
        (
            admonition_node.source,
            admonition_node.line,
        ) = self.state_machine.get_source_and_line(self.lineno)
        if isinstance(admonition_node, sphinx.ext.todo.todo_node):
            # todo admonitions need extra info for the todolist directive
            admonition_node["docname"] = admonition_node.source
            self.state.document.note_explicit_target(admonition_node)
        else:
            self.add_name(admonition_node)
        if "collapsible" in self.options:
            admonition_node["collapsible"] = self.options["collapsible"]
        if "no-title" in self.options and "collapsible" in self.options:
            logger.error(
                "title is needed for collapsible admonitions",
                location=admonition_node,
            )
            del self.options["no-title"]  # force-disable option
        textnodes, messages = self.state.inline_text(title_text, self.lineno)
        if "no-title" not in self.options and title_text:
            title = nodes.title(title_text, "", *textnodes)
            title.source, title.line = self.state_machine.get_source_and_line(
                self.lineno
            )
            admonition_node += title
        admonition_node += messages
        admonition_node["classes"] += self.classes
        self.state.nested_parse(self.content, self.content_offset, admonition_node)
        return [admonition_node]


def get_directive_class(name, title, classes=None) -> Type[CustomAdmonitionDirective]:
    """A helper function to produce a admonition directive's class."""
    # alias upstream-deprecated CSS classes for pre-defined admonitions in sphinx
    class_list = [nodes.make_id(name)]
    if classes:
        class_list.extend(classes)

    # uncomment this block when we merge v9.x from upstream
    # if name in ("caution", "attention"):
    #     class_list.append("warning")
    # elif name == "error":
    #     class_list.append("danger")
    # elif name in ("important", "hint"):
    #     class_list.append("tip")
    # elif name == "todo":
    #     class_list.append("info")

    class CustomizedAdmonition(CustomAdmonitionDirective):
        default_title = title
        classes = class_list
        optional_arguments = int(name not in admonitionlabels)
        node_class = nodes.admonition if name != "todo" else sphinx.ext.todo.todo_node

    return CustomizedAdmonition


def on_builder_inited(app: Sphinx):
    """register the directives for the custom admonitions and build the CSS."""
    config = app.config
    custom_admonitions: List[CustomAdmonitionConfig] = getattr(
        config, "sphinx_immaterial_custom_admonitions"
    )
    setattr(app.builder.env, "sphinx_immaterial_custom_icons", {})
    for admonition in custom_admonitions:

        app.add_directive(
            name=admonition.name,
            cls=get_directive_class(
                admonition.name, admonition.title, admonition.classes
            ),
            override=admonition.override,
        )

        # set variables for CSS template to match HTML output from generated directives
        admonition.name = nodes.make_id(admonition.name)
        if admonition.icon is not None:
            admonition.icon = load_svg_into_builder_env(app.builder, admonition.icon)
    setattr(app.builder.env, "sphinx_immaterial_custom_admonitions", custom_admonitions)


def on_config_inited(app: Sphinx, config: Config):
    """Add admonitions based on CSS classes inherited from mkdocs-material theme."""

    # override the generic admonition directive
    if getattr(config, "sphinx_immaterial_override_generic_admonitions"):
        app.add_directive("admonition", get_directive_class("admonition", ""), True)

    # generate directives for inherited admonitions from upstream CSS
    if getattr(config, "sphinx_immaterial_generate_extra_admonitions"):
        for admonition in (
            "abstract",
            "info",
            "success",
            "question",
            "failure",
            "bug",
            "example",
            "quote",
        ):
            app.add_directive(
                admonition,
                get_directive_class(admonition, _(admonition.title())),
            )
    confval_name = "sphinx_immaterial_custom_admonitions"
    # validate user defined config for custom admonitions
    user_defined_admonitions = pydantic.parse_obj_as(
        List[CustomAdmonitionConfig], getattr(config, confval_name)
    )
    setattr(config, confval_name, user_defined_admonitions)
    user_defined_dir_names = [directive.name for directive in user_defined_admonitions]

    # override the specific admonitions defined in sphinx and docutils
    # these are the admonitions that have translated titles in sphinx.locale
    if getattr(config, "sphinx_immaterial_override_builtin_admonitions"):
        for admonition, title in admonitionlabels.items():

            if admonition in user_defined_dir_names:
                continue
            app.add_directive(admonition, get_directive_class(admonition, title), True)


def consolidate_css(app: Sphinx, env: BuildEnvironment):
    """Generates the CSS for icons and admonitions, then appends that to the
    theme's bundled CSS."""

    theme_options = app.config["html_theme_options"]
    is_palette_defined = False
    if theme_options:
        is_palette_defined = "palette" in theme_options

    custom_admonitions = getattr(env, "sphinx_immaterial_custom_admonitions")
    custom_icons = getattr(env, "sphinx_immaterial_custom_icons")
    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(str(PurePath(__file__).parent))
    )
    template = jinja_env.get_template("custom_admonitions.css")
    generated = template.render(
        icons=custom_icons,
        admonitions=custom_admonitions,
    )

    css_data = io.BytesIO()
    # copy pre-minified CSS from theme's bundles
    theme_bundles = Path(__file__).parent / "bundles" / "stylesheets"
    for bundle in theme_bundles.glob("*.min.css"):
        if not is_palette_defined and bundle.name.startswith("palette"):
            continue  # don't need the palette CSS if HTML doesn't need it
        css_data.write(bundle.read_bytes())

    # append the generated CSS for icons and admonitions
    css_data.write(generated.replace("\n", "").encode(encoding="utf-8"))

    # hash CSS data and write to generated file
    css_data_hash = hashlib.sha256(css_data.getvalue()).hexdigest()
    css_name = f"stylesheets/sphinx_immaterial_theme.{css_data_hash[:17]}.min.css"
    css_output = Path(app.outdir, "_static", css_name)
    css_output.parent.mkdir(parents=True, exist_ok=True)
    css_output.write_bytes(css_data.getvalue())

    # ensure new CSS file is added in the rendered HTML output
    app.add_css_file(css_name)


def setup(app: Sphinx):
    """register our custom directive."""
    app.add_config_value(
        name="sphinx_immaterial_custom_admonitions",
        default=[],
        rebuild="env",
        types=[List[CustomAdmonitionConfig]],
    )
    app.add_config_value(
        name="sphinx_immaterial_generate_extra_admonitions",
        default=True,
        rebuild="html",
        types=bool,
    )
    app.add_config_value(
        name="sphinx_immaterial_override_generic_admonitions",
        default=True,
        rebuild="html",
        types=bool,
    )
    app.add_config_value(
        name="sphinx_immaterial_override_builtin_admonitions",
        default=True,
        rebuild="html",
        types=bool,
    )

    app.connect("builder-inited", on_builder_inited)
    app.connect("env-check-consistency", consolidate_css)
    app.connect("config-inited", on_config_inited)

    patch_visit_admonition()
    patch_depart_admonition()

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Converts a preprocessed C++ source file into a JSON API description.

This program expects as input preprocessed C++ source code obtained using the
`-E` (preprocessed), `-C` (preserve comments), and `-dD` (preserve macro
definitions) GCC/Clang options.  It also accepts a list of compiler options to
be passed to Clang, although these are mostly irrelevant due to the prior
preprocessing.

It uses a combination of the libclang Python bindings and the C++ parser in the
Sphinx Python package to parse declarations and definitions and produce a JSON
representation of the API.

The JSON representation describes the API as a set of "entities", each with a
unique id, and relations between them.

Libclang can handle arbitrary C++ syntax, but provides very restricted access to
the resultant AST.  To workaround those limitations, in some cases extracted
declarations or portions of declarations are converted back to source
representation and re-parsed using the C++ parser provided by the Sphinx C++
domain.  The C++ parser in Sphinx is extremely limited and as it does not rely
on a symbol table does not always correctly handle template arguments, but
provides full access to its AST.

This program performs a number of transformations on the declarations:

- Exported entities are filtered using various criteria

- `std::enable_if_t` uses are converted to C++20 requires clauses.

- Internal return types are elided (replaced with auto).
"""

import argparse
import collections
import dataclasses
import enum
import functools
import json
import os
import pathlib
import re
import sys
import time
import typing
from typing import (
    cast,
    Dict,
    Any,
    List,
    Tuple,
    Optional,
    Sequence,
    Union,
    ClassVar,
    Pattern,
)
from textwrap import dedent

import ctypes

import clang.cindex
from clang.cindex import (
    Cursor,
    CursorKind,
    Token,
    TokenKind,
    TranslationUnit,
    SourceLocation,
    SourceRange,
)
import docutils.nodes
import pydantic.dataclasses
import sphinx.domains.cpp
import sphinx.util.logging

from . import ast_fixes  # type: ignore[unused-import]

logger = sphinx.util.logging.getLogger(__name__)

_UNMATCHABLE_REGEXP = re.compile("a^")


def _combine_regexp_list(items: Sequence[Union[str, Pattern[str]]]) -> re.Pattern:
    if not items:
        return _UNMATCHABLE_REGEXP

    def get_parenthesized_source(x: Union[str, Pattern[str]]):
        if isinstance(x, re.Pattern):
            x = x.pattern
        return f"(?:{x})"

    return re.compile("|".join(get_parenthesized_source(x) for x in items))


def _make_replacement_pattern(
    strings: List[str], prefix: str, suffix: str
) -> re.Pattern:
    if not strings:
        return _UNMATCHABLE_REGEXP
    return re.compile(
        "|".join(rf"(?:{prefix}{re.escape(before)}{suffix})" for before in strings)
    )


TEMPLATE_PARAMETER_ENABLE_IF_TYPE_PATTERN = re.compile(
    r"\s*(?:typename|class)\s*=\s*std\s*::\s*enable_if_t\s*<(.*)>\s*"
)
TEMPLATE_PARAMETER_ENABLE_IF_NON_TYPE_PATTERN = re.compile(
    r"\s*std\s*::\s*enable_if_t\s*<(.*)>\s*\*\s*=\s*(nullptr|0)\s*"
)

SPECIAL_GROUP_COMMAND_PATTERN = re.compile(
    r"^(?:\\|@)(ingroup|relates|membergroup|id)\s+(.+[^\s])\s*$", re.MULTILINE
)


@pydantic.dataclasses.dataclass
class Config:
    """Specifies a C++ API parsing configuration.

    Based on this configuration, a description of the API is generated.
    """

    input_path: str = "__input.cpp"
    """Path to the input file to parse.

    This may either be a path to an existing file, or `.input_content` may
    specify its content, in which case the filesystem is not accessed.

    If `.input_content` is specified and merely contains :cpp:`#include`
    directives, then the actual path does not matter and may be left as the
    default value.
    """

    input_content: Optional[bytes] = None
    """Specifies the content of `.input_path`.

    If unspecified, the content is read from filesystem.
    """

    compiler_flags: List[str] = dataclasses.field(default_factory=list)
    """List of compiler flags to pass to Clang."""

    verbose: bool = False
    """Parse in verbose mode."""

    include_directory_map: Dict[str, str] = dataclasses.field(default_factory=dict)
    """Maps actual include directories to a displayed directory name.

    The keys should be prefixes of paths specified in error messages/source
    locations identified by clang.

    The values should be the corresponding prefix to use in the documented
    :cpp:`#include` paths.
    """

    allow_paths: List[Pattern] = dataclasses.field(
        default_factory=lambda: [re.compile("")]
    )
    """List of regular expressions matching *allowed* paths.

    Only entities defined in files that match `.allow_paths`, and don't match
    `.disallow_paths`, are documented.  By default all entities are documented,
    but this default is not normally usable, because it will include entities
    defined in the standard library and third-party libraries.

    .. important::
        When building on Windows, all path separators are normalized to :python:`"/"`.
        Therefore, in the specified regular expressions, always use :python:`"/"` to
        match a path separator.
    """

    disallow_paths: List[Pattern] = dataclasses.field(default_factory=list)
    """List of regular expressions matching *disallowed* paths.

    Entities defined in files matching any of these patterns are not documented.

    .. important::
        When building on Windows, all path separators are normalized to :python:`"/"`.
        Therefore, in the specified regular expressions, always use :python:`"/"` to
        match a path separator.
    """

    disallow_namespaces: List[Pattern] = dataclasses.field(default_factory=list)
    """List of regular expressions matching *disallowed* namespaces.

    Entities defined in namespaces matching any of the specified patterns are
    not documented.
    """

    allow_symbols: List[Pattern] = dataclasses.field(
        default_factory=lambda: [re.compile("")]
    )
    """List of regular expressions matching *allowed* symbols.

    Only symbols matching one of the `.allow_symbols` patterns, and not matching
    `.disallow_symbols`, are documented.  By default, all symbols are allowed.
    """

    disallow_symbols: List[Pattern] = dataclasses.field(default_factory=list)
    """List of regular expressions matching *disallowed* symbols.

    Symbols matching any of these patterns are undocumented.
    """

    allow_macros: List[Pattern] = dataclasses.field(
        default_factory=lambda: [re.compile("")]
    )
    """List of regular expressions matching *allowed* macros.

    Only macros names matching `.allow_macros`, and not matching
    `.disallow_macros`, are documented.
    """

    disallow_macros: List[Pattern] = dataclasses.field(default_factory=list)
    """List of regular expressions matching *disallowed* macro names.

    Macros matching any of these patterns are undocumented.
    """

    ignore_diagnostics: List[Pattern] = dataclasses.field(default_factory=list)
    """List of regular expressions matching diagnostics to ignore.

    Diagnostics matching any of these patterns are ignored.
    """

    template_parameter_enable_if_patterns: List[Pattern] = dataclasses.field(
        default_factory=lambda: [
            TEMPLATE_PARAMETER_ENABLE_IF_TYPE_PATTERN,
            TEMPLATE_PARAMETER_ENABLE_IF_NON_TYPE_PATTERN,
        ]
    )
    type_replacements: Dict[str, str] = dataclasses.field(default_factory=dict)
    hide_types: List[Pattern] = dataclasses.field(default_factory=list)
    """List of regular expressions matching *hidden* types.

    Matching return types are replaced with :cpp:`auto`, and matching
    initializers are elided.
    """
    ignore_template_parameters: List[Pattern] = dataclasses.field(default_factory=list)
    """List of regular expressions matching *ignored* template parameters.

    Template parameters with a declaration matching any of these patterns are
    excluded from the generated documentation.
    """

    hide_initializers: List[Pattern] = dataclasses.field(
        default_factory=lambda: [re.compile(r"^=\s*(?:(true|false)\s*$|\[)")]
    )
    """List of regular expressions matching initializers to elide.

    Any matching initializer expression is elided from the generated
    documentation.
    """

    # Derived from `allow_paths`.
    allow_path_pattern: ClassVar[Pattern]

    # Derived from `disallow_paths`.
    disallow_path_pattern: ClassVar[Pattern]

    # Derived from `allow_symbols`.
    allow_symbols_pattern: ClassVar[Pattern]

    # Derived from `disallow_symbols`.
    disallow_symbols_pattern: ClassVar[Pattern]

    # Derived from `allow_macros`.
    allow_macros_pattern: ClassVar[Pattern]

    # Derived from `disallow_macros`.
    disallow_macros_pattern: ClassVar[Pattern]

    # Derived from `ignore_diagnostics`.
    ignore_diagnostics_pattern: ClassVar[Pattern]

    # Derived from `hide_types`.
    hide_types_pattern: ClassVar[Pattern]

    # Derived from `type_replacements`.
    type_replacements_pattern: ClassVar[Pattern]

    # Derived from `ignore_template_parameters`.
    ignore_template_parameters_pattern: ClassVar[Pattern]

    # Derived from `hide_initializers`.
    hide_initializers_pattern: ClassVar[Pattern]

    include_directory_map_pattern: ClassVar[Pattern]

    def __post_init__(self):
        self.allow_path_pattern = _combine_regexp_list(self.allow_paths)
        self.disallow_path_pattern = _combine_regexp_list(self.disallow_paths)
        self.allow_path_pattern = _combine_regexp_list(self.allow_paths)
        self.disallow_namespaces_pattern = _combine_regexp_list(
            self.disallow_namespaces
        )
        self.allow_symbols_pattern = _combine_regexp_list(self.allow_symbols)
        self.disallow_symbols_pattern = _combine_regexp_list(self.disallow_symbols)
        self.allow_macros_pattern = _combine_regexp_list(self.allow_macros)
        self.disallow_macros_pattern = _combine_regexp_list(self.disallow_macros)
        self.ignore_diagnostics_pattern = _combine_regexp_list(self.ignore_diagnostics)
        self.hide_types_pattern = _combine_regexp_list(self.hide_types)
        self.type_replacements_pattern = _make_replacement_pattern(
            self.type_replacements, prefix=r"\b", suffix=r"\b"
        )
        self.ignore_template_parameters_pattern = _combine_regexp_list(
            self.ignore_template_parameters
        )
        self.hide_initializers_pattern = _combine_regexp_list(self.hide_initializers)
        if os.name == "nt":
            self._include_directory_map = {
                key.replace("\\", "/"): value
                for key, value in self.include_directory_map.items()
            }
        else:
            self._include_directory_map = self.include_directory_map
        self.include_directory_map_pattern = _make_replacement_pattern(
            list(self._include_directory_map.keys()), prefix="^", suffix=""
        )
        self._cached_mapped_include_directories = {}

    _include_directory_map: ClassVar[Dict[str, str]]
    _cached_mapped_include_directories: ClassVar[Dict[str, str]]

    def map_include_path(self, path: str) -> str:
        mapped = self._cached_mapped_include_directories.get(path)
        if mapped is not None:
            return mapped
        if os.name == "nt":
            path = path.replace("\\", "/")
        if path.startswith("./"):
            path = path[2:]
        new_mapped = self.include_directory_map_pattern.sub(
            lambda m: self._include_directory_map[m.group(0)], path
        )
        self._cached_mapped_include_directories[path] = new_mapped
        return new_mapped


CppApiEntity = Dict[str, Any]
EntityId = str


def get_entity_id(cursor: Cursor) -> EntityId:
    # USR workarounds from:
    # https://github.com/foonathan/cppast/blob/e558e2d58f519e3a83af770d460672b1d4ba2886/src/libclang/parse_functions.cpp#L13
    usr = cursor.get_usr()
    if cursor.kind in (CursorKind.FUNCTION_TEMPLATE, CursorKind.CONVERSION_FUNCTION):
        # Combine return type with USR to prevent collisions
        return f"{usr} {cursor.result_type.spelling}"
    if cursor.kind == CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION:
        # libclang issue: templ<T()> vs templ<T() &>
        # but identical USR
        # same workaround: combine display name with usr
        # (and hope this prevents all collisions...)
        return f"{usr} {cursor.displayname}"
    return usr


def _substitute_internal_type_names(config: Config, decl: str) -> str:
    return config.type_replacements_pattern.sub(
        lambda m: config.type_replacements[m.group(0)], decl
    )


def get_previous_line_location(tu, location: SourceLocation):
    f = location.file
    line = location.line
    return SourceLocation.from_position(tu, location.file, line - 1, 1)


def get_presumed_location(location: SourceLocation) -> typing.Tuple[str, int, int]:
    f, l, c = clang.cindex._CXString(), ctypes.c_uint(), ctypes.c_uint()
    clang.cindex.conf.lib.clang_getPresumedLocation(
        location, ctypes.byref(f), ctypes.byref(l), ctypes.byref(c)
    )
    return (clang.cindex._CXString.from_result(f), int(l.value), int(c.value))


def _get_template_cursor_kind(cursor: Cursor) -> CursorKind:
    return CursorKind.from_id(clang.cindex.conf.lib.clang_getTemplateCursorKind(cursor))


def _get_specialized_cursor_template(cursor: Cursor) -> typing.Optional[Cursor]:
    return clang.cindex.conf.lib.clang_getSpecializedCursorTemplate(cursor)


def _is_doc_comment(token: Token):
    return token.spelling.startswith("///")


def _get_full_nested_name(cursor: typing.Optional[Cursor]) -> str:
    if cursor is None:
        return ""
    ancestors = []
    while True:
        if cursor.kind == CursorKind.TRANSLATION_UNIT:
            break
        if cursor.kind == CursorKind.NAMESPACE:
            name = cursor.spelling
        else:
            name = cursor.displayname
        ancestors.append(name + "::")
        cursor = cursor.semantic_parent
    ancestors.reverse()
    return "".join(ancestors)


CLASS_KINDS = (
    CursorKind.STRUCT_DECL,
    CursorKind.CLASS_DECL,
    CursorKind.CLASS_TEMPLATE,
    CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION,
)


def _get_all_decls(config: Config, cursor: Cursor, allow_file):

    NAMESPACE = CursorKind.NAMESPACE
    for child in cursor.get_children():
        location = child.location
        if location.file is None:
            continue
        kind = child.kind
        if kind == NAMESPACE:
            if (
                not allow_file or allow_file(get_presumed_location(location)[0])
            ) and not config.disallow_namespaces_pattern.match(child.spelling):
                yield from _get_all_decls(config, child, None)
            continue
        if kind not in ALLOWED_KINDS:
            continue
        if allow_file and not allow_file(get_presumed_location(location)[0]):
            continue
        if child.kind == CursorKind.MACRO_DEFINITION:
            yield child
            continue
        yield child
        if kind in CLASS_KINDS:
            yield from _get_all_decls(config, child, None)


def split_doc_comment_into_lines(cmt: str) -> List[str]:
    """Strip the raw string of an object's comment into lines.
    :param cmt: the comment to parse.
    :returns: A list of the lines without the surrounding C++ comment syntax.
    """
    # split into a list of lines & account for CRLF and LF line endings
    body = [line.rstrip("\r") for line in cmt.splitlines()]

    # strip all the comment syntax out
    if body[0].startswith("//"):
        body = [line.lstrip("/").lstrip("!").lstrip("<") for line in body]
    elif body[0].startswith("/*"):
        body[0] = body[0].lstrip("/").lstrip("*").lstrip("!")
        multi_lined_asterisk = True  # works also for single-line comments blocks
        if len(body) > 1:
            line_has_asterisk = [line.startswith("*") for line in body[1:]]
            multi_lined_asterisk = line_has_asterisk.count(True) == len(body) - 1
        body = [
            (line.lstrip("*").lstrip("<") if multi_lined_asterisk else line)
            for line in body
        ]
        body[-1] = body[-1].rstrip("*/").rstrip()
    body = dedent("\n".join(body)).splitlines()
    return [""] if not body else body


NON_DOC_COMMENT = re.compile(
    r"(^//[^/\!].*$\n)|(^/\*[^\*\!](?:.|\n)*?\*/$\n)", re.MULTILINE
)


def get_doc_comment(config: Config, cursor: Cursor):
    translation_unit = cursor.translation_unit
    for token in cursor.get_tokens():
        location = token.location
        break
    else:
        location = cursor.location
    f = location.file
    line = location.line
    end_location = SourceLocation.from_position(translation_unit, f, line, 1)
    comment = cursor.raw_comment
    if not comment:
        return None
    comment_lines = []
    # The first line is never indented (in `raw_comment` form).
    # Clang doesn't strip indentation from subsequent lines in an indented block.
    # So, dedent all subsequent lines only
    first_line_end = comment.find("\n")
    comment = comment[:first_line_end] + dedent(comment[first_line_end:])

    # remove any non-docstring comments
    match = NON_DOC_COMMENT.search(comment)
    while match is not None:
        # strip comment syntax from the block before non-doc comment
        comment_lines.extend(split_doc_comment_into_lines(comment[: match.start()]))
        # Append blank lines as replacement of non-doc comment.
        # This should retain the src's line numbers
        comment_lines.extend(["\n"] * match.group(0).count("\n"))
        comment = comment[match.end() :]
        match = NON_DOC_COMMENT.search(comment)
    if comment:  # strip comment from any block that remains after non-doc comment
        comment_lines.extend(split_doc_comment_into_lines(comment))
    return {
        "text": "\n".join(comment_lines),
        "location": _get_location_json(config, end_location),
    }


class Extractor:
    def __init__(self, config: Config):
        self.config = config

        input_path = config.input_path

        input_content = config.input_content
        if input_content is None:
            input_content = pathlib.Path(input_path).read_bytes()

        input_content = re.sub(
            b"#pragma clang module", b"//#pragma clang module", input_content
        )

        self.input_source = input_content

        self.index = clang.cindex.Index.create()
        start_time = time.time()
        self.tu = self.index.parse(
            input_path,
            unsaved_files=[(input_path, input_content)],
            args=tuple(config.compiler_flags) + ("-ferror-limit=0",),
            options=(  # TranslationUnit.PARSE_SKIP_FUNCTION_BODIES +
                TranslationUnit.PARSE_DETAILED_PROCESSING_RECORD
            ),
        )
        end_time = time.time()
        if config.verbose:
            logger.info("Parsed C++ input in %.5f seconds", end_time - start_time)

        for diag in self.tu.diagnostics:
            if config.ignore_diagnostics_pattern.search(diag.spelling):
                if config.verbose:
                    logger.info(
                        diag.spelling,
                        location=_get_location_string(config, diag.location),
                    )
                continue
            logger.error(
                diag.spelling, location=_get_location_string(config, diag.location)
            )

        def _allow_file(path: str) -> bool:
            path = config.map_include_path(path)
            if not config.allow_path_pattern.search(path):
                return False
            if config.disallow_path_pattern.search(path):
                return False
            return True

        self.decls = list(
            _get_all_decls(
                config, self.tu.cursor, functools.lru_cache(maxsize=None)(_allow_file)
            )
        )


EXCLUDED_COMPILER_FLAGS = frozenset(
    [
        "-Xclang=-disable-noundef-analysis",
    ]
)


def _transform_type_alias_decl(config: Config, decl: Cursor):
    underlying_type: Optional[str] = _substitute_internal_type_names(
        config, decl.underlying_typedef_type.spelling
    )
    assert underlying_type is not None
    if config.hide_types_pattern.search(underlying_type):
        underlying_type = None

    return {
        "kind": "alias",
        "name": decl.spelling,
        "underlying_type": underlying_type,
    }


def get_extent_spelling(translation_unit: TranslationUnit, extent: SourceRange) -> str:
    """Returns the C++ source representation for the specified extent.

    Comments are excluded and for simplicity all tokens are separated by
    whitespace.  This results in excessive whitespace, but that does not matter
    because this is intended to be parsed by the Sphinx cpp domain anyway.
    """

    def get_spellings():
        prev_token = None
        COMMENT = TokenKind.COMMENT
        spellings = []
        for token in translation_unit.get_tokens(extent=extent):
            if prev_token is not None:
                yield prev_token.spelling
                prev_token = None
            if token.kind == COMMENT:
                continue
            prev_token = token
        # We need to handle the last token specially, because clang sometimes parses
        # ">>" as a single token but the extent may cover only the first of the two
        # angle brackets.
        if prev_token is not None:
            spelling = prev_token.spelling
            token_end = prev_token.extent.end
            offset_diff = token_end.offset - extent.end.offset
            if offset_diff != 0:
                yield spelling[:-offset_diff]
            else:
                yield spelling

    return " ".join(get_spellings())


def get_related_comments(decl: Cursor):
    # casts below are workaround for: https://github.com/tgockel/types-clang/pull/2
    start = cast(SourceLocation, decl.extent.start)
    end = cast(SourceLocation, decl.extent.end)
    # Move forward one line to avoid skipping any comments on the last line
    tu = decl.translation_unit
    end = SourceLocation.from_position(tu, end.file, end.line + 1, 1)
    COMMENT = TokenKind.COMMENT
    for token in tu.get_tokens(
        extent=SourceRange.from_locations(cast(int, start), cast(int, end))
    ):
        if token.kind != COMMENT:
            continue
        yield token


NONITPICK_PATTERN = re.compile(r"//\s*NONITPICK:\s*(.*[^\s])\s*")


def get_nonitpick_directives(decl: Cursor) -> List[str]:
    directives = []
    for comment in get_related_comments(decl):
        text = comment.spelling
        m = NONITPICK_PATTERN.match(text)
        if m is None:
            continue
        directives.append(m.group(1))
    return directives


TEMPLATE_CURSOR_KINDS = frozenset(
    [
        CursorKind.FUNCTION_TEMPLATE,
        CursorKind.CLASS_TEMPLATE,
        CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION,
        CursorKind.TYPE_ALIAS_TEMPLATE_DECL,
    ]
)

TEMPLATE_PARAMETER_KIND_TO_JSON_KIND = {
    CursorKind.TEMPLATE_TYPE_PARAMETER: "type",
    CursorKind.TEMPLATE_NON_TYPE_PARAMETER: "non_type",
    CursorKind.TEMPLATE_TEMPLATE_PARAMETER: "template",
}


def _clang_template_parameter_to_json(config: Config, decl: Cursor):
    param_decl_str = get_extent_spelling(decl.translation_unit, decl.extent)
    param = _parse_template_parameter(param_decl_str)
    if param is None:
        return {
            "declaration": param_decl_str,
            "name": decl.spelling,
            "kind": TEMPLATE_PARAMETER_KIND_TO_JSON_KIND[decl.kind],
            # Heuristic to determine if it is a pack.
            "pack": "..." in param_decl_str,
        }
    return _sphinx_ast_template_parameter_to_json(config, param)


def _get_template_parameters(config: Config, decl: Cursor):
    if decl.kind not in TEMPLATE_CURSOR_KINDS:
        return None
    result = []
    for child in decl.get_children():
        if child.kind not in (
            CursorKind.TEMPLATE_TYPE_PARAMETER,
            CursorKind.TEMPLATE_NON_TYPE_PARAMETER,
            CursorKind.TEMPLATE_TEMPLATE_PARAMETER,
        ):
            continue
        result.append(_clang_template_parameter_to_json(config, child))
    return result


def _get_non_template_kind(cursor: Cursor):
    kind = cursor.kind
    if kind not in TEMPLATE_CURSOR_KINDS:
        return kind
    return _get_template_cursor_kind(cursor)


def _transform_type_alias_template_decl(config: Config, decl: Cursor):
    underlying_type: Optional[str]
    for child in decl.get_children():
        if child.kind == CursorKind.TYPE_ALIAS_DECL:
            underlying_type = _substitute_internal_type_names(
                config, child.underlying_typedef_type.spelling
            )
            break
    else:
        raise ValueError("Could not determine underlying type")

    requires = []

    if re.search(r"^\s*std\s*::\s*enable_if_t\s*<", underlying_type) is not None:
        presumed_file, presumed_line, _ = get_presumed_location(decl.location)
        parser = sphinx.domains.cpp.DefinitionParser(
            underlying_type,
            location=(presumed_file, presumed_line),
            config=cast(sphinx.config.Config, SphinxConfig()),
        )
        ast = parser._parse_type(False)
        parser.skip_ws()
        parser.assert_end()
        assert isinstance(ast, sphinx.domains.cpp.ASTType)
        requires_expr = _extract_requires_from_enable_if_t_type(config, ast)
        if requires_expr is not None:
            requires.append(requires_expr)
            underlying_type = str(ast)

    if config.hide_types_pattern.search(underlying_type) is not None:
        underlying_type = None
    return {
        "kind": "alias",
        "name": decl.spelling,
        "underlying_type": underlying_type,
        "requires": requires,
    }


def _get_class_keyword(kind: CursorKind) -> str:
    return "class" if kind == CursorKind.CLASS_DECL else "struct"


def _get_bases(config: Config, decl: Cursor):
    for child in decl.get_children():
        if child.kind != CursorKind.CXX_BASE_SPECIFIER:
            continue
        type_spelling = _substitute_internal_type_names(config, child.type.spelling)
        if config.hide_types_pattern.search(type_spelling) is not None:
            continue
        yield {"type": type_spelling, "access": child.access_specifier.name.lower()}


def _transform_class_decl(config: Config, decl: Cursor):
    obj = {
        "kind": "class",
        "keyword": _get_class_keyword(decl.kind),
        "name": decl.displayname,
        "prefix": _parse_declaration_prefix(decl, is_class=True),
        "bases": list(_get_bases(config, decl)),
    }
    specializes = _get_specialized_cursor_template(decl)
    if specializes:
        obj["specializes"] = get_entity_id(specializes)
    return obj


def _transform_class_template_decl(config: Config, decl: Cursor):
    return {
        "kind": "class",
        "keyword": _get_class_keyword(_get_template_cursor_kind(decl)),
        "name": decl.spelling,
        "prefix": _parse_declaration_prefix(decl, is_class=True),
        "bases": list(_get_bases(config, decl)),
    }


def _transform_class_template_partial_specialization_decl(config: Config, decl: Cursor):
    return {
        "kind": "class",
        "keyword": _get_class_keyword(_get_template_cursor_kind(decl)),
        "name": decl.displayname,
        "specializes": get_entity_id(
            cast(Cursor, _get_specialized_cursor_template(decl))
        ),
        "prefix": _parse_declaration_prefix(decl, is_class=True),
        "bases": list(_get_bases(config, decl)),
    }


def _get_function_parameters(decl: Cursor):
    if decl.kind == CursorKind.FUNCTION_DECL:
        yield from decl.get_arguments()
        return
    for child in decl.get_children():
        if child.kind != CursorKind.PARM_DECL:
            continue
        yield child


FUNCTION_CURSOR_KIND_TO_JSON_KIND = {
    CursorKind.FUNCTION_DECL: "function",
    CursorKind.CXX_METHOD: "method",
    CursorKind.CONSTRUCTOR: "constructor",
    CursorKind.DESTRUCTOR: "destructor",
    CursorKind.CONVERSION_FUNCTION: "conversion_function",
}


def _parse_declaration_prefix(decl: Cursor, is_class: bool) -> typing.List[str]:
    decl_extent = decl.extent
    start_location = decl_extent.start
    end_location = None
    prefix_parts = []
    for child in decl.get_children():
        # Skip template introduction
        if child.kind in (
            CursorKind.TEMPLATE_TYPE_PARAMETER,
            CursorKind.TEMPLATE_NON_TYPE_PARAMETER,
            CursorKind.TEMPLATE_TEMPLATE_PARAMETER,
        ):
            start_location = child.extent.end
            continue
        if child.kind.is_attribute():
            attr_spelling = get_extent_spelling(decl.translation_unit, child.extent)
            prefix_parts.append(f"[[{attr_spelling}]]")
            continue
        end_location = child.extent.start
        break

    if not is_class:
        for token in decl.translation_unit.get_tokens(
            extent=SourceRange.from_locations(
                start_location, end_location or decl_extent.end
            )
        ):
            # skip `inline` since that is not an important part of the API
            if token.spelling in ("explicit", "constexpr"):
                prefix_parts.append(token.spelling)
    return prefix_parts


def _get_declaration_spelling(decl: Cursor):
    decl_extent = decl.extent
    start_location = decl_extent.start
    end_location = None
    for child in decl.get_children():
        if child.kind.is_statement():
            end_location = child.extent.start
            break
    else:
        end_location = decl_extent.end
    return get_extent_spelling(
        decl.translation_unit,
        extent=SourceRange.from_locations(start_location, end_location),
    )


def _transform_function_decl(config: Config, decl: Cursor):
    name = decl.spelling
    if name.startswith("<deduction guide for "):
        # Exclude deduction guides for now
        return None
    non_template_kind = decl.kind
    if decl.kind == CursorKind.FUNCTION_TEMPLATE:
        non_template_kind = _get_template_cursor_kind(decl)

    specializes = _get_specialized_cursor_template(decl)

    if non_template_kind == CursorKind.CONSTRUCTOR:
        # TODO: handle = default, = delete
        first_bracket = name.find("<")
        if first_bracket != -1:
            name = name[:first_bracket]

        prefix = _parse_declaration_prefix(decl, is_class=False)
        if decl.storage_class == clang.cindex.StorageClass.STATIC:
            prefix.insert(0, "static")
        source_code = _get_declaration_spelling(decl)
        name_substitute = _pick_name_substitute(source_code)
        decl_string = (
            "".join(x + " " for x in prefix)
            + name_substitute
            + "("
            + ", ".join(
                get_extent_spelling(decl.translation_unit, arg.extent)
                for arg in _get_function_parameters(decl)
            )
            + ")"
        )
        requires_expr = None
    else:
        (
            decl_string,
            bare_name,
            template_args,
            name_substitute,
            requires_expr,
        ) = _parse_function(config, decl)
        name = bare_name
        if specializes and template_args is not None:
            name += template_args

    arity = sum(x.kind == CursorKind.PARM_DECL for x in decl.get_children())

    obj = {
        "kind": FUNCTION_CURSOR_KIND_TO_JSON_KIND[non_template_kind],
        "name": name,
        "arity": arity,
        "declaration": decl_string,
        "name_substitute": name_substitute,
        "requires": [requires_expr] if requires_expr else None,
    }
    if specializes:
        obj["specializes"] = get_entity_id(specializes)
    return obj


def _transform_enum_decl(config: Config, decl: Cursor):
    keyword = None
    tokens = list(decl.get_tokens())
    assert len(tokens) >= 2
    assert tokens[0].spelling == "enum"
    token1_spelling = tokens[1].spelling
    if token1_spelling in ("class", "struct"):
        keyword = token1_spelling

    name = decl.spelling
    enumerators = []
    for child in decl.get_children():
        if child.kind != CursorKind.ENUM_CONSTANT_DECL:
            continue
        enumerators.append(
            {
                "kind": "enumerator",
                "id": get_entity_id(child),
                "name": child.spelling,
                "decl": get_extent_spelling(decl.translation_unit, child.extent),
                "doc": get_doc_comment(config, child),
                "location": _get_location_json(config, child.location),
            }
        )
    return {
        "kind": "enum",
        "keyword": keyword,
        "name": decl.spelling,
        "enumerators": enumerators,
    }


def _pick_name_substitute(code: str) -> str:
    i = 0
    while True:
        substitute = f"__x{i}"
        if substitute not in code:
            return substitute
        i += 1


def _transform_var_decl(config: Config, decl: Cursor):
    exprs = [x for x in decl.get_children() if x.kind.is_expression()]
    presumed_filename, presumed_line, _ = get_presumed_location(decl.location)
    if len(exprs) > 1:
        raise ValueError(
            "%s:%d: Expected VAR decl to have at most one expression as a child: %r, but has: %d"
            % (presumed_filename, presumed_line, decl.spelling, len(exprs))
        )

    prefix = _parse_declaration_prefix(decl, is_class=False)

    type_spelling = decl.type.spelling
    if "(lambda at " in type_spelling:
        type_spelling = "auto"
    name_substitute = _pick_name_substitute(type_spelling)

    initializer = None
    if len(exprs) == 1:
        initializer = "= " + get_extent_spelling(
            decl.translation_unit, exprs[0].extent
        ).rstrip(";")
        if _is_internal_initializer(config, initializer):
            initializer = None
    declaration = " ".join(prefix) + " " + type_spelling + " " + name_substitute
    return {
        "kind": "var",
        "name": decl.spelling,
        "declaration": declaration,
        "name_substitute": name_substitute,
        "initializer": initializer,
    }


class SphinxConfig:

    cpp_id_attributes: Any = []
    cpp_paren_attributes: Any = []


def _parse_name(name: str, template_prefix: str) -> sphinx.domains.cpp.ASTNestedName:
    parser = sphinx.domains.cpp.DefinitionParser(
        f"{template_prefix} int {name}",
        location=("", 0),
        config=cast(sphinx.config.Config, SphinxConfig()),
    )
    ast = parser.parse_declaration("member", "member")
    parser.skip_ws()
    parser.assert_end(allowSemicolon=True)
    return ast.name


def _substitute_name(
    top_ast: sphinx.domains.cpp.ASTDeclaration,
    ast: sphinx.domains.cpp.ASTType,
    source_code: str,
) -> str:
    name_substitute = _pick_name_substitute(source_code)
    template_args = ast.name.names[-1].templateArgs
    name_substitute_with_args = name_substitute
    if template_args is not None:
        name_substitute_with_args += str(template_args)

    template_prefix = ""
    if top_ast.templatePrefix is not None:
        template_prefix = str(top_ast.templatePrefix.templates[-1])

    ast.name = _parse_name(name_substitute_with_args, template_prefix=template_prefix)
    return name_substitute_with_args


def _maybe_wrap_requires_expr_in_parentheses(expr: str) -> str:
    parser = sphinx.domains.cpp.DefinitionParser(
        "requires " + expr,
        location=("", 0),
        config=cast(sphinx.config.Config, SphinxConfig()),
    )
    try:
        parser._parse_requires_clause()
        parser.skip_ws()
        parser.assert_end()
        return expr
    except:  # pylint: disable=bare-except
        return f"({expr})"


def _extract_requires_from_enable_if_t_type(
    config: Config, ast: sphinx.domains.cpp.ASTType
) -> typing.Optional[str]:
    if not isinstance(
        ast.declSpecs.trailingTypeSpec, sphinx.domains.cpp.ASTTrailingTypeSpecName
    ):
        return None
    decl_specs = ast.declSpecs
    trailing_type_spec = decl_specs.trailingTypeSpec
    if not str(trailing_type_spec).startswith("std::enable_if_t<"):
        return None
    template_args = trailing_type_spec.name.names[1].templateArgs.args  # type: ignore[attr-defined]
    requires_expr = str(template_args[0])
    if len(template_args) == 2:
        result_type = str(template_args[1])
    else:
        result_type = "void"

    parser = sphinx.domains.cpp.DefinitionParser(
        result_type, location=("", 0), config=cast(sphinx.config.Config, SphinxConfig())
    )
    new_ast = parser._parse_type(False)
    parser.skip_ws()
    parser.assert_end()

    new_decl_specs = new_ast.declSpecs

    def copy_qualifiers(
        orig_d: sphinx.domains.cpp.ASTDeclarator,
        new_d: sphinx.domains.cpp.ASTDeclarator,
    ):
        if isinstance(new_d, sphinx.domains.cpp.ASTDeclaratorRef):
            return sphinx.domains.cpp.ASTDeclaratorRef(
                next=copy_qualifiers(orig_d, new_d.next), attrs=new_d.attrs
            )
        if isinstance(new_d, sphinx.domains.cpp.ASTDeclaratorPtr):
            return sphinx.domains.cpp.ASTDeclaratorPtr(
                next=copy_qualifiers(orig_d, new_d.next),
                volatile=new_d.volatile,
                const=new_d.const,
                attrs=new_d.attrs,
            )
        return orig_d

    ast.decl = copy_qualifiers(ast.decl, new_ast.decl)

    decl_specs.trailingTypeSpec = new_decl_specs.trailingTypeSpec
    decl_specs.leftSpecs.const = (
        decl_specs.leftSpecs.const or new_decl_specs.leftSpecs.const
    )
    decl_specs.leftSpecs.volatile = (
        decl_specs.leftSpecs.volatile or new_decl_specs.leftSpecs.volatile
    )
    decl_specs.rightSpecs.const = (
        decl_specs.rightSpecs.const or new_decl_specs.rightSpecs.const
    )
    decl_specs.rightSpecs.volatile = (
        decl_specs.rightSpecs.volatile or new_decl_specs.rightSpecs.volatile
    )

    return _substitute_internal_type_names(config, requires_expr)


_FUNCTION_NAME_REPLACEMENTS = {
    "operator[ ]": "operator[]",
    "operator( )": "operator()",
}


def _parse_function(config: Config, decl: Cursor):

    presumed_file, presumed_line, _ = get_presumed_location(decl.location)
    source_code = _get_declaration_spelling(decl)
    parser = sphinx.domains.cpp.DefinitionParser(
        source_code,
        location=(presumed_file, presumed_line),
        config=cast(sphinx.config.Config, SphinxConfig()),
    )
    ast = parser.parse_declaration("function", "function")
    parser.skip_ws()
    parser.assert_end(allowSemicolon=True)
    assert isinstance(ast.declaration, sphinx.domains.cpp.ASTType)

    requires_expr = _extract_requires_from_enable_if_t_type(config, ast.declaration)

    last_name_element = ast.declaration.name.names[-1]

    bare_name = str(last_name_element.identOrOp)
    bare_name = _FUNCTION_NAME_REPLACEMENTS.get(bare_name, bare_name)
    template_args = last_name_element.templateArgs

    template_args_str = str(template_args) if template_args is not None else None

    name_substitute = _substitute_name(ast, ast.declaration, source_code)

    # Exclude `inline` specifier since it isn't really informative in API
    # documentation.
    ast.declaration.declSpecs.leftSpecs.inline = False

    decl_string = _substitute_internal_type_names(config, str(ast.declaration))
    return decl_string, bare_name, template_args_str, name_substitute, requires_expr


def _is_internal_initializer(config: Config, initializer: str) -> bool:
    return (
        config.hide_initializers_pattern.search(initializer) is not None
        or config.hide_types_pattern.search(initializer) is not None
    )


def _sphinx_ast_template_parameter_to_json(
    config: Config, param: sphinx.domains.cpp.ASTTemplateParam
) -> Any:
    if isinstance(param, sphinx.domains.cpp.ASTTemplateParamType):
        kind = "type"
    elif isinstance(param, sphinx.domains.cpp.ASTTemplateParamTemplateType):
        kind = "template"
    else:
        kind = "non_type"

    return {
        "declaration": _substitute_internal_type_names(config, str(param)),
        "name": str(param.get_identifier()),
        "kind": kind,
        "pack": param.isPack,  # type: ignore[attr-defined]
    }


def _transform_unexposed_decl(config: Config, decl: Cursor):
    # libclang unfortunately does not support variable templates; they are only
    # exposed as an unexposed decl.

    source_code = get_extent_spelling(decl.translation_unit, decl.extent)

    # Note: Since `source_code` is reconstructed from the tokens, we don't need to
    # worry about inconsistency in spacing.
    if not source_code.startswith("template <"):
        return None

    # Assume that it is a variable template
    # Attempt to parse it via sphinx's c++ domain parser

    presumed_file, presumed_line, _ = get_presumed_location(decl.location)

    try:
        parser = sphinx.domains.cpp.DefinitionParser(
            source_code,
            location=(presumed_file, presumed_line),
            config=cast(sphinx.config.Config, SphinxConfig()),
        )
        ast = parser.parse_declaration("member", "member")
        parser.skip_ws()
        parser.assert_end(allowSemicolon=True)

        declaration = cast(
            Union[
                sphinx.domains.cpp.ASTTypeWithInit,
                sphinx.domains.cpp.ASTTemplateParamConstrainedTypeWithInit,
            ],
            ast.declaration,
        )
        template_args = declaration.type.name.names[-1].templateArgs
        name = str(declaration.type.name.names[-1])
        name_substitute = _substitute_name(ast, declaration.type, source_code)

        decl_string = _substitute_internal_type_names(config, str(declaration.type))
        decl_string = re.sub("(^| )inline ", " ", decl_string)

        initializer: Optional[str] = _substitute_internal_type_names(
            config, str(declaration.init).strip().rstrip(";").strip()
        )
        assert initializer is not None
        if _is_internal_initializer(config, initializer):
            initializer = None

        obj: Dict[str, Any] = {
            "kind": "var",
            "name": name,
            "template_parameters": [
                _sphinx_ast_template_parameter_to_json(
                    config, cast(sphinx.domains.cpp.ASTTemplateParam, t)
                )
                for t in cast(
                    sphinx.domains.cpp.ASTTemplateDeclarationPrefix, ast.templatePrefix
                )
                .templates[-1]
                .params
            ],
            "declaration": decl_string,
            "name_substitute": name_substitute,
            "initializer": initializer,
        }
        if template_args is not None:
            obj["specializes"] = True
        return obj

    except Exception as e:
        raise ValueError("Failed to parse unexposed") from e
    return None


def _parse_macro_parameters(decl: Cursor) -> typing.Optional[typing.List[str]]:
    # Check if the macro is a function-like macro
    # `cast` below is workaround for: https://github.com/tgockel/types-clang/pull/2
    token_iterator = cast(typing.Iterator[Token], decl.get_tokens())
    first_token = next(token_iterator)
    assert first_token.spelling == decl.spelling
    def_start_offset = cast(SourceLocation, first_token.extent.end).offset
    try:
        next_token = next(token_iterator)
    except StopIteration:
        return None
    if next_token.spelling != "(":
        return None
    if next_token.location.offset != def_start_offset:
        # There is a space between the macro name and the first "(", which means
        # this is not a function-like macro.
        return None

    parameters = []
    for token in token_iterator:
        if token.kind == TokenKind.COMMENT:
            continue
        spelling = token.spelling
        if spelling == ")":
            break
        if spelling == ",":
            continue
        parameters.append(spelling)
    else:
        presumed_file, presumed_line, _ = get_presumed_location(decl.location)
        raise ValueError(
            "%s:%d: Failed to parse macro %s"
            % (presumed_file, presumed_line, decl.spelling)
        )
    return parameters


def _transform_macro(config: Config, decl: Cursor):
    name = decl.spelling
    if config.disallow_macros_pattern.search(name) is not None:
        return None
    return {
        "kind": "macro",
        "name": name,
        "parameters": _parse_macro_parameters(decl),
    }


TRANSFORMERS = {
    CursorKind.TYPE_ALIAS_DECL: _transform_type_alias_decl,
    CursorKind.TYPEDEF_DECL: _transform_type_alias_decl,
    CursorKind.TYPE_ALIAS_TEMPLATE_DECL: _transform_type_alias_template_decl,
    CursorKind.STRUCT_DECL: _transform_class_decl,
    CursorKind.CLASS_DECL: _transform_class_decl,
    CursorKind.CLASS_TEMPLATE: _transform_class_template_decl,
    CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION: _transform_class_template_partial_specialization_decl,
    CursorKind.FUNCTION_DECL: _transform_function_decl,
    CursorKind.CXX_METHOD: _transform_function_decl,
    CursorKind.CONVERSION_FUNCTION: _transform_function_decl,
    CursorKind.CONSTRUCTOR: _transform_function_decl,
    CursorKind.DESTRUCTOR: _transform_function_decl,
    CursorKind.FUNCTION_TEMPLATE: _transform_function_decl,
    CursorKind.ENUM_DECL: _transform_enum_decl,
    CursorKind.VAR_DECL: _transform_var_decl,
    CursorKind.FIELD_DECL: _transform_var_decl,
    CursorKind.UNEXPOSED_DECL: _transform_unexposed_decl,
    CursorKind.MACRO_DEFINITION: _transform_macro,
}

ALLOWED_KINDS = frozenset(list(TRANSFORMERS.keys()) + [CursorKind.FRIEND_DECL])


def _parse_args(output_required: bool):
    ap = argparse.ArgumentParser()
    ap.add_argument("--config", type=str, required=True)
    ap.add_argument("--output", type=str, required=output_required)
    return ap.parse_args()


def _merge_decl_json(existing_json, new_json):
    if existing_json["doc"] and new_json["doc"]:
        raise ValueError("Duplicate doc strings: %r and %r" % (existing_json, new_json))
    existing_json["doc"] = existing_json["doc"] or new_json["doc"]
    template_parameters = existing_json.get("template_parameters")
    if template_parameters:
        new_template_parameters = new_json.get("template_parameters")
        for i, old_param in enumerate(template_parameters):
            new_param = new_template_parameters[i]
            if new_param.startswith(old_param):
                template_parameters[i] = new_param
            elif not old_param.startswith(new_param):
                raise ValueError(
                    "Conflicting template parameter %d: %r and %r"
                    % (i, existing_json, new_json)
                )


def _get_location_json(config: Config, location: SourceLocation):
    filename, line, col = get_presumed_location(location)
    filename = config.map_include_path(filename)
    return {"file": filename, "line": line - 1}


def _get_location_string(config: Config, location: SourceLocation):
    filename, line, col = get_presumed_location(location)
    filename = config.map_include_path(filename)
    return f"{filename}:{line}:{col}"


def _is_immediately_after(decl: Cursor, prev_decl: Cursor) -> bool:
    # casts below are workaround for: https://github.com/tgockel/types-clang/pull/2
    prev_end = cast(SourceLocation, prev_decl.extent.end)
    cur_start = cast(SourceLocation, decl.extent.start)
    cur_file, cur_line, _ = get_presumed_location(cur_start)
    prev_file, prev_line, _ = get_presumed_location(prev_end)
    return cur_file == prev_file and cur_line == prev_line + 1


_NORMALIZED_KIND = {
    "constructor": "function",
    "conversion_function": "function",
    "method": "function",
    "function": "function",
    "class": "class",
    "alias": "alias",
    "enum": "enum",
    "macro": "macro",
    "var": "var",
}


def _kinds_are_compatible(a: str, b: str) -> bool:
    return _NORMALIZED_KIND[a] == _NORMALIZED_KIND[b]


class JsonApiGenerator:
    def __init__(self, extractor):
        self.extractor = extractor
        self.config = extractor.config
        self.seen_decls = {}
        self.output_json = []
        self._prev_decl = None
        self._document_with_parent = {}

    def _resolve_document_with(self, entity_id: EntityId) -> EntityId:
        while True:
            document_with_parent = self._document_with_parent.get(entity_id)
            if document_with_parent is None:
                break
            entity_id = document_with_parent
        return entity_id

    def _transform_cursor_to_json(self, decl: Cursor, parent: Optional[Cursor]):
        doc = get_doc_comment(self.config, decl)
        document_with = None
        location = _get_location_json(self.config, decl.location)
        if not doc:
            if self._prev_decl is not None and _is_immediately_after(
                decl, self._prev_decl[0]
            ):
                document_with = self._resolve_document_with(self._prev_decl[1]["id"])
            else:
                # Exclude undocumented entities
                return None
        else:
            if (
                self._prev_decl is not None
                and self._prev_decl[1]["location"] == location
            ):
                # Same line as previous declaration, presumably due to macro expansion
                # generating multiple declarations.
                #
                # Document as a sibling of the previous declaration.
                document_with = self._resolve_document_with(self._prev_decl[1]["id"])
        transformer = TRANSFORMERS.get(decl.kind)
        if transformer is None:
            return None
        json_repr = transformer(self.config, decl)
        if json_repr is None:
            return None
        if parent is None or parent.kind in (
            CursorKind.NAMESPACE,
            CursorKind.TRANSLATION_UNIT,
        ):
            json_repr["scope"] = _get_full_nested_name(parent)
        else:
            json_repr["parent"] = get_entity_id(parent)
        entity_id = get_entity_id(decl)
        if document_with:
            prev_json = self._prev_decl[1]
            if (
                prev_json is None
                or not _kinds_are_compatible(prev_json["kind"], json_repr["kind"])
                or prev_json.get("parent") != json_repr.get("parent")
                or prev_json.get("scope") != json_repr.get("scope")
            ):
                if not doc:
                    # Undocumented and can't document with previous decl
                    return None
                document_with = None
            if document_with is not None:
                doc = None
                self._document_with_parent[entity_id] = document_with
                json_repr["document_with"] = document_with
        extent = decl.extent
        json_repr["location"] = location
        nonitpick = get_nonitpick_directives(decl)
        if nonitpick:
            json_repr["nonitpick"] = nonitpick
        json_repr["doc"] = doc
        if decl.kind != CursorKind.UNEXPOSED_DECL:
            template_parameters = _get_template_parameters(self.config, decl)
            if json_repr.get("specializes") and template_parameters is None:
                template_parameters = []
            json_repr["template_parameters"] = template_parameters
        json_repr["id"] = entity_id
        return json_repr

    def add(self, decl: Cursor):
        is_friend = False
        if decl.kind == CursorKind.FRIEND_DECL:
            # Check if this is a hidden friend function.
            children = list(decl.get_children())
            if len(children) != 1:
                return
            decl = children[0]
            if not decl.kind.is_declaration():
                return
            is_friend = True
            parent = decl.lexical_parent
        else:
            parent = decl.semantic_parent
        json_repr = self._transform_cursor_to_json(decl, parent)
        if json_repr is None:
            self._prev_decl = None
            return
        json_repr["friend"] = is_friend
        parent_id = json_repr.get("parent")
        if parent_id is not None and parent_id not in self.seen_decls:
            # Parent is undocumented, skip.
            return
        self._prev_decl = (decl, json_repr)
        entity_id = json_repr["id"]
        existing_json_repr = self.seen_decls.get(entity_id)
        if existing_json_repr is not None:
            _merge_decl_json(existing_json_repr, json_repr)
            return
        self.seen_decls[entity_id] = json_repr


def _parse_template_parameter(
    decl: str,
) -> Optional[sphinx.domains.cpp.ASTTemplateParam]:
    # Note: We must include an extra trailing ">" because
    # `_parse_template_parameter` fails if the parameter is not followed by "," or
    # ">".
    parser = sphinx.domains.cpp.DefinitionParser(
        decl + ">", location=("", 0), config=cast(sphinx.config.Config, SphinxConfig())
    )
    parser.allowFallbackExpressionParsing = False
    try:
        param = parser._parse_template_parameter()
        assert parser.skip_string(">")
        parser.assert_end()
        return param
    except sphinx.domains.cpp.DefinitionError:
        return None


def _extract_sfinae_replacement(template_parameter: str) -> Optional[Tuple[str, str]]:
    param = _parse_template_parameter(template_parameter)
    if param is None:
        return None

    name = str(param.get_identifier())

    if not name.lower().startswith("sfinae"):
        return None

    if isinstance(param, sphinx.domains.cpp.ASTTemplateParamType):
        default_type = param.data.default
        if default_type is None:
            return None
        return (name, str(default_type))

    if isinstance(param, sphinx.domains.cpp.ASTTemplateParamNonType):
        default_value: Optional[sphinx.domains.cpp.ASTBase] = param.param.init
        if default_value is None:
            return None
        if isinstance(default_value, sphinx.domains.cpp.ASTInitializer):
            default_value = default_value.value
        return (name, str(default_value))

    return None


CONDITIONALLY_EXPLICIT_PATTERN = re.compile(r"\(ExplicitRequires\((.+)\)\)")


def _match_template_parameter_enable_if_pattern(
    config: Config, decl: str
) -> Optional[str]:
    for pattern in config.template_parameter_enable_if_patterns:
        m = pattern.fullmatch(decl)
        if m is not None:
            return m.group(1)
    return None


def _transform_template_parameters(config: Config, template_parameters: List[Any]):
    """Transforms template parameters to C++20 form."""

    requires = []
    new_template_parameters = []
    replacements: Dict[str, str] = {}

    for template_parameter in template_parameters:
        decl = template_parameter["declaration"]
        requires_expr = _match_template_parameter_enable_if_pattern(config, decl)
        if requires_expr is not None:
            requires.append(requires_expr)
            continue
        if config.ignore_template_parameters_pattern.fullmatch(decl):
            continue
        # If the template parameter is of the form `YYY SfinaeXXX = Condition`, then
        # we want to exclude it from the template parameter list and instead return
        # the substitution `{"SfinaeXXX": "Condition"}`.  To avoid parsing in cases
        # that can't possibly match, first look to see if the name starts with
        # `"sfinae"`.
        if re.match(r"sfinae", template_parameter["name"], re.IGNORECASE) is not None:
            # Possibly match, parse to verify.
            replacement = _extract_sfinae_replacement(decl)
            if replacement is not None:
                replacements[replacement[0]] = replacement[1]
                continue
        new_template_parameters.append(template_parameter)

    return (
        new_template_parameters,
        requires,
        replacements,
    )


def _strip_return_type(
    declaration: str, template_prefix: str, location: Tuple[str, int]
) -> str:
    parser = sphinx.domains.cpp.DefinitionParser(
        template_prefix + declaration,
        location=location,
        config=cast(sphinx.config.Config, SphinxConfig()),
    )
    ast = parser.parse_declaration("function", "function")
    parser.skip_ws()
    parser.assert_end()
    assert isinstance(ast.declaration, sphinx.domains.cpp.ASTType)
    ast.declaration.declSpecs.trailingTypeSpec = (
        sphinx.domains.cpp.ASTTrailingTypeSpecFundamental(["auto"], ["auto"])
    )
    return str(ast.declaration)


_OPERATOR_PAGE_NAMES = {
    ("operator+", 1): "operator-unary_plus",
    ("operator-", 1): "operator-negate",
    ("operator*", 1): "operator-dereference",
    ("operator~", 1): "operator-complement",
    ("operator!", 1): "operator-logical_not",
    ("operator++", 1): "operator-pre_inc",
    ("operator++", 2): "operator-post_inc",
    ("operator--", 1): "operator-pre_dec",
    ("operator--", 2): "operator-post_dec",
    ("operator<<", 2): "operator-shift_left",
    ("operator>>", 2): "operator-shift_right",
    ("operator+", 2): "operator-plus",
    ("operator-", 2): "operator-minus",
    ("operator*", 2): "operator-multiplies",
    ("operator/", 2): "operator-divides",
    ("operator%", 2): "operator-modulus",
    ("operator<", 2): "operator-less",
    ("operator<=", 2): "operator-less_equal",
    ("operator>=", 2): "operator-greater_equal",
    ("operator==", 2): "operator-equal_to",
    ("operator!=", 2): "operator-not_equal_to",
    ("operator=", 2): "operator-assign",
    ("operator<<=", 2): "operator-shift_left_assign",
    ("operator>>=", 2): "operator-shift_right_assign",
    ("operator*=", 2): "operator-multiplies_assign",
    ("operator/=", 2): "operator-divides_assign",
    ("operator%=", 2): "operator-modulus_assign",
    ("operator+=", 2): "operator-plus_assign",
    ("operator-=", 2): "operator-minus_assign",
    ("operator&=", 2): "operator-bitwise_and_assign",
    ("operator|=", 2): "operator-bitwise_or_assign",
    ("operator&&", 2): "operator-logical_and",
    ("operator||", 2): "operator-logical_or",
    ("operator|", 2): "operator-bitwise_or",
    ("operator&", 2): "operator-bitwise_and",
    ("operator^", 2): "operator-bitwise_xor",
    ("operator,", 2): "operator-comma",
    ("operator->", 1): "operator-arrow",
    ("operator&", 1): "operator-address_of",
    "operator()": "operator-call",
    "operator[]": "operator-subscript",
}

DEFAULT_MEMBER_GROUP_FOR_MEMBER_ENTITY_TYPE = {
    "constructor": "Constructors",
    "destructor": "Constructors",
    "class": "Types",
    "method": "Methods",
    "enum": "Types",
    "alias": "Types",
    "conversion_function": "Conversion operators",
    "var": "Data members",
    "function": "Friend functions",
}

DEFAULT_MEMBER_GROUP_FOR_NON_MEMBER_ENTITY_TYPE = {
    "alias": "Related Types",
    "enum": "Related Types",
    "class": "Related Types",
    "function": "Related Functions",
    "var": "Related Constants",
    "macro": "Related Macros",
}


def _get_default_member_group(entity: CppApiEntity) -> str:
    if entity.get("parent"):
        return DEFAULT_MEMBER_GROUP_FOR_MEMBER_ENTITY_TYPE[entity["kind"]]
    return DEFAULT_MEMBER_GROUP_FOR_NON_MEMBER_ENTITY_TYPE[entity["kind"]]


def _normalize_doc_text(text: str) -> str:
    text = re.sub(r"^((?:\\|@)(?:brief|details)\s+)", "", text, flags=re.MULTILINE)
    text = re.sub(
        r"^(?:\\|@)(t?param)(\[(?:in|out|in,\sout)\])?\s+([a-zA-Z_][^ ]*)",
        ":\\1 \\3\\2:",
        text,
        flags=re.MULTILINE,
    )
    text = re.sub(
        r"^(?:\\|@)(error)\s+`([^`]+)`", ":\\1 \\2:", text, flags=re.MULTILINE
    )
    text = re.sub(
        r"^(?:\\|@)(returns?|pre|post|[ds]?checks|invariant|requires)(?: |\n )",
        ":\\1: ",
        text,
        flags=re.MULTILINE,
    )
    text = re.sub(r"^(?:\\|@)(retval)\s+(\S+)", ":\\1 \\2:", text, flags=re.MULTILINE)
    text = SPECIAL_GROUP_COMMAND_PATTERN.sub("", text)
    return text


FUNCTION_ENTITY_KINDS = frozenset(
    ["function", "method", "constructor", "destructor", "conversion_function"]
)


def _is_function(entity: CppApiEntity) -> bool:
    return entity["kind"] in FUNCTION_ENTITY_KINDS


def _get_path_component_from_special_id(special_id: str) -> str:
    special_id = re.sub("[^a-zA-Z0-9_]+", "-", special_id)
    return special_id.strip("-")


def _apply_identifier_replacements(
    requires_term: str, replacements: Dict[str, str]
) -> str:
    for orig_identifier, replacement in replacements.items():
        requires_term = re.sub(
            r"\b" + re.escape(orig_identifier) + r"\b", replacement, requires_term
        )
    return requires_term


def _make_explicit_conditional(decl: str, explicit: str) -> str:
    new_str = re.sub(r"\bexplicit\b", f"explicit({explicit})", decl, 1)
    if new_str == decl:
        raise ValueError(
            "Failed to insert explicit condition %r into: %s"
            % (
                explicit,
                decl,
            )
        )
    return new_str


def _is_uniform_binary_expr(
    expr: sphinx.domains.cpp.ASTBase, allowed_ops: Tuple[str, ...]
) -> bool:
    if not isinstance(expr, sphinx.domains.cpp.ASTBinOpExpr):
        return False
    return all(op in allowed_ops for op in expr.ops)


def _is_logical_and_expr(expr: sphinx.domains.cpp.ASTBase) -> bool:
    return _is_uniform_binary_expr(expr, ("&&", "and"))


def _is_primary_expr(expr: sphinx.domains.cpp.ASTBase) -> bool:
    return isinstance(
        expr,
        (
            sphinx.domains.cpp.ASTLiteral,
            sphinx.domains.cpp.ASTIdExpression,
            sphinx.domains.cpp.ASTThisLiteral,
        ),
    )


def _normalize_requires_terms(terms: List[str]) -> List[str]:
    if not terms:
        return terms
    expr = " && ".join(f"({term})" for term in terms)
    parser = sphinx.domains.cpp.DefinitionParser(
        "requires " + expr,
        location=("", 0),
        config=cast(sphinx.config.Config, SphinxConfig()),
    )
    # If we allow fallback parsing, we end up with incorrect parsing and a
    # spurious warning.
    parser.allowFallbackExpressionParsing = False
    ast = parser._parse_requires_clause()
    parser.skip_ws()
    parser.assert_end()

    assert ast is not None

    new_terms = []

    def process(
        expr: Union[sphinx.domains.cpp.ASTType, sphinx.domains.cpp.ASTExpression]
    ):
        while True:
            if isinstance(expr, sphinx.domains.cpp.ASTParenExpr):
                expr = expr.expr
                continue
            if isinstance(expr, sphinx.domains.cpp.ASTBinOpExpr) and not expr.ops:
                expr = expr.exprs[0]
                continue
            if (
                isinstance(expr, sphinx.domains.cpp.ASTPostfixExpr)
                and not expr.postFixes
            ):
                expr = expr.prefix
                continue
            break
        if _is_logical_and_expr(expr):
            for sub_expr in cast(sphinx.domains.cpp.ASTBinOpExpr, expr).exprs:
                process(sub_expr)
            return
        if _is_primary_expr(expr):
            new_terms.append(str(expr))
            return
        new_terms.append(f"({expr})")

    process(ast.expr)

    return new_terms


def _format_template_arguments(entity: CppApiEntity) -> str:
    if entity.get("specializes"):
        # Template arguments already included in `entity["name"]`.
        return ""
    template_parameters = entity.get("template_parameters")
    if not template_parameters:
        return ""
    strs = []
    for param in template_parameters:
        arg = param["name"]
        if not arg:
            continue
        if param["pack"]:
            arg += "..."
        strs.append(arg)
    args_str = ", ".join(strs)
    return f"<{args_str}>"


def _get_entity_base_page_name_component(entity: CppApiEntity) -> str:
    base_name = entity["name"]
    if entity["kind"] == "class" and entity.get("specializes"):
        # Strip any template arguments
        base_name = re.sub("([^<]*).*", r"\1", base_name)
    elif entity["kind"] == "conversion_function":
        base_name = "operator-cast"
    elif entity["kind"] in ("function", "method") and re.match(
        r"operator\b", base_name
    ):
        arity = entity["arity"]
        if entity["kind"] == "method":
            arity += 1
        op_page_name = _OPERATOR_PAGE_NAMES.get((base_name, arity))
        if op_page_name is None:
            op_page_name = _OPERATOR_PAGE_NAMES[base_name]
        base_name = op_page_name

    return base_name


def _get_entity_page_name_component(entity: CppApiEntity) -> str:
    page_name = _get_entity_base_page_name_component(entity)
    special_id = entity.get("special_id")
    if special_id is not None:
        page_name += f"-{_get_path_component_from_special_id(special_id)}"
    return page_name


def _ensure_unique_page_names(
    entities_with_page_names: List[EntityId],
    entities: Dict[EntityId, CppApiEntity],
    warning,
) -> None:

    names: Dict[
        Tuple[Optional[str], Optional[str], str, Optional[str]], List[EntityId]
    ] = {}

    for entity_id in entities_with_page_names:
        entity = entities[entity_id]
        parent_id = entity.get("parent")
        special_id = entity.get("special_id")
        scope = entity.get("scope")
        base_name = _get_entity_base_page_name_component(entity)
        key = (parent_id, scope, base_name, special_id)
        names.setdefault(key, []).append(entity_id)

    for (parent_id, scope, base_name, special_id), entity_ids in names.items():
        if len(entity_ids) == 1:
            continue
        page_entities = [entities[entity_id] for entity_id in entity_ids]
        warning(
            "Disambiguating %d overloads of %s using numerical ids.  Definitions at %s",
            len(entity_ids),
            base_name,
            ", ".join(
                "%s:%d" % (entity["location"]["file"], entity["location"]["line"])
                for entity in page_entities
            ),
        )

        for i, entity in enumerate(page_entities):
            entity["special_id"] = str(i + 1)


def organize_entities(config: Config, entities: Dict[EntityId, CppApiEntity]):

    errors: List[Tuple[str, Optional[Tuple[str, int]]]] = []
    warnings: List[Tuple[str, Optional[Tuple[str, int]]]] = []

    def error(msg: str, *args, location=None):
        loc = (location["file"], location["line"]) if location is not None else None
        errors.append((msg % args, loc))

    def warning(msg: str, *args, location=None):
        loc = (location["file"], location["line"]) if location is not None else None
        warnings.append((msg % args, loc))

    def _handle_document_with(entity: CppApiEntity) -> bool:
        document_with = entity.get("document_with")
        if document_with is None:
            return False
        sibling_entity: Optional[CppApiEntity] = entities.get(document_with)
        if sibling_entity is None:
            return False
        sibling_entity.setdefault("siblings", []).append(entity["id"])
        return True

    def _normalize_entity_requires(entity: CppApiEntity):
        template_parameters = entity.get("template_parameters")
        if template_parameters:
            (
                template_parameters,
                requires,
                replacements,
            ) = _transform_template_parameters(config, template_parameters)
            if entity.get("specializes") is None and not template_parameters:
                entity["template_parameters"] = None
            else:
                entity["template_parameters"] = template_parameters
        else:
            requires = []
            replacements = None
            explicit = None

        existing_requires = entity.get("requires")
        if existing_requires:
            requires = existing_requires + requires

        if _is_function(entity):
            declaration = entity["declaration"]
            if replacements:
                declaration = _apply_identifier_replacements(declaration, replacements)
            if (
                entity["kind"] != "constructor"
                and config.hide_types_pattern.search(
                    declaration[: declaration.index(entity["name_substitute"])]
                )
                is not None
            ):
                declaration = _strip_return_type(
                    declaration,
                    "template <> " if template_parameters is not None else "",
                    location=(entity["location"]["file"], entity["location"]["line"]),
                )
            entity["declaration"] = declaration
        else:
            if replacements:
                for key in ("declaration", "underlying_type"):
                    x = entity.get(key, None)
                    if x is not None:
                        entity[key] = _apply_identifier_replacements(x, replacements)

        if replacements:
            requires = [
                _apply_identifier_replacements(x, replacements) for x in requires
            ]

        requires = _normalize_requires_terms(requires)
        new_requires = []
        explicit = None
        for term in requires:
            m = CONDITIONALLY_EXPLICIT_PATTERN.fullmatch(term)
            if m is not None:
                if explicit is not None:
                    raise ValueError(
                        "cannot have more than one conditionally-explicit term"
                    )
                explicit = m.group(1)
            else:
                new_requires.append(term)
        requires = new_requires

        if explicit:
            if entity["kind"] != "constructor":
                raise ValueError(
                    "conditionally-explicit terms only valid on constructors"
                )
            entity["declaration"] = _make_explicit_conditional(
                entity["declaration"], explicit
            )

        requires = [x for x in requires if config.hide_types_pattern.search(x) is None]
        entity["requires"] = requires

    def get_entity_page_name(entity: CppApiEntity) -> str:
        components = []
        cur_entity = entity
        while True:
            components.append(_get_entity_page_name_component(cur_entity))
            parent_id = cur_entity.get("parent")
            if parent_id is None:
                break
            parent_entity = entities.get(parent_id)
            assert parent_entity is not None
            cur_entity = parent_entity
        components.reverse()
        page_name = (cur_entity["scope"] + "::".join(components)).replace("::", ".")
        return page_name

    def _parse_entity_doc(entity: CppApiEntity):
        doc = entity["doc"]
        if doc is None:
            if _handle_document_with(entity):
                return True
            return False
        doc_text = doc["text"]
        for m in SPECIAL_GROUP_COMMAND_PATTERN.finditer(doc_text):
            entity["special_" + m.group(1)] = m.group(2).strip()
        return True

    def get_entity_scope(entity: CppApiEntity) -> str:
        components = []
        cur_entity = entity
        while True:
            parent_id = cur_entity.get("parent")
            if parent_id is None:
                break
            parent_entity = entities.get(parent_id)
            assert parent_entity is not None
            cur_entity = parent_entity
            name_with_args = cur_entity["name"]
            if not cur_entity.get("specializes"):
                name_with_args += _format_template_arguments(cur_entity)
            components.append(name_with_args)
        components.reverse()
        if components:
            components.append("")
        return cur_entity.get("scope", "") + "::".join(components)

    def get_entity_object_name(entity: CppApiEntity) -> str:
        name = get_entity_scope(entity) + entity["name"]
        special_id = entity.get("special_id")
        if special_id:
            name += f"[{special_id}]"
        return name

    unspecialized_names: Dict[
        Tuple[Optional[EntityId], Optional[str], str], EntityId
    ] = {}

    names: Dict[str, EntityId] = {}

    def resolve_entity_name(
        scope: str, relative_entity_name: str
    ) -> Optional[EntityId]:
        if relative_entity_name.startswith("::"):
            resolved = relative_entity_name[2:]
            entity_id = names.get(resolved)
            if entity_id is None:
                return None
            return entity_id
        truncate_idx = len(scope)
        while True:
            full_name = scope[:truncate_idx] + relative_entity_name
            entity_id = names.get(full_name)
            if entity_id is not None:
                return entity_id
            if truncate_idx == 0:
                return None
            truncate_idx = scope.rfind("::", 0, truncate_idx - 2)
            if truncate_idx == -1:
                truncate_idx = 0
            else:
                truncate_idx = truncate_idx + 2

    must_resolve_specializes: List[CppApiEntity] = []

    all_nonitpick = []

    def _handle_nitpick(entity: CppApiEntity, targets: List[str]) -> None:
        document_with = entity.get("document_with")
        if document_with:
            entity = entities[document_with]
        location = entity["location"]
        filename = location["file"]
        line = location["line"]
        for target in targets:
            all_nonitpick.append({"file": filename, "line": line, "target": target})

    entities_with_page_names: List[EntityId] = []

    for entity in entities.values():
        specializes = entity.get("specializes")
        if (
            entity["kind"] == "var"
            and entity.get("template_parameters") is not None
            and specializes is None
        ):
            key = (entity.get("parent"), entity.get("scope"), entity["name"])
            entity_id = entity["id"]
            if unspecialized_names.setdefault(key, entity_id) != entity_id:
                raise ValueError("Duplicate unspecialized entity name: %r" % (key,))
        if specializes is True:
            must_resolve_specializes.append(entity)
        if not _parse_entity_doc(entity):
            continue
        _normalize_entity_requires(entity)
        nonitpick = entity.get("nonitpick")
        if nonitpick:
            _handle_nitpick(entity, nonitpick)
        if not entity["doc"]:
            continue
        entities_with_page_names.append(entity["id"])

    for entity in must_resolve_specializes:
        name = cast(str, entity["name"])
        name = name[: name.index("<")]
        unspecialized_key = (entity.get("parent"), entity.get("scope"), name)
        unspecialized_id = unspecialized_names.get(unspecialized_key)
        if unspecialized_id is not None:
            entity["specializes"] = unspecialized_id

    _ensure_unique_page_names(entities_with_page_names, entities, warning)
    for entity_id in entities_with_page_names:
        entity = entities[entity_id]
        names[get_entity_object_name(entity)] = entity_id
        entity["page_name"] = get_entity_page_name(entity)
        doc: Any = entity["doc"]
        doc["text"] = _normalize_doc_text(doc["text"])

    groups: Dict[str, List[EntityId]] = {}

    for entity in entities.values():
        entity_id = entity["id"]
        doc = entity["doc"]
        if doc is None:
            continue
        ingroup = entity.get("special_ingroup")
        relates_name = entity.get("special_relates")
        member_group = entity.get("special_membergroup")

        if (ingroup is not None) and (relates_name is not None):
            error(
                "Cannot specify both \\ingroup and \\relates for %r",
                entity,
                location=doc["location"],
            )
            continue

        if ingroup is not None:
            ingroup = docutils.nodes.make_id(ingroup)
            groups.setdefault(ingroup, []).append(entity_id)
            if member_group is not None:
                error(
                    "Cannot specify both \\ingroup and \\membergroup for %r",
                    entity,
                    location=doc["location"],
                )
            continue

        parent_id = entity.get("parent")
        if relates_name is not None:
            scope = get_entity_scope(entity)
            relates_id = resolve_entity_name(scope, relates_name)
            if relates_id is None:
                error(
                    "Cannot resolve entity reference %r from scope %r",
                    relates_name,
                    scope,
                    location=doc["location"],
                )
                continue
            parent_id = None
        else:
            if parent_id is None:
                warning(
                    "No group or relates specified for entity %s%s",
                    entity.get("scope"),
                    entity["name"],
                    location=doc["location"],
                )
                continue
            relates_id = parent_id

        if member_group is None:
            member_group = _get_default_member_group(entity)
        assert relates_id is not None
        cast(
            dict,
            entities[relates_id].setdefault(
                "related_members" if parent_id is not None else "related_nonmembers", {}
            ),
        ).setdefault(member_group, []).append(entity_id)

    return {
        "entities": entities,
        "groups": groups,
        "errors": errors,
        "warnings": warnings,
        "nonitpick": all_nonitpick,
    }


def _get_output_json(extractor: Extractor):
    generator = JsonApiGenerator(extractor)
    if extractor.config.verbose:
        logger.info("Found %d C++ declarations", len(extractor.decls))
    for decl in extractor.decls:
        generator.add(decl)
    return organize_entities(extractor.config, generator.seen_decls)


def generate_output(config: Config):
    extractor = Extractor(config)
    return _get_output_json(extractor)


def _load_config(config_path: str) -> Config:

    config_content = pathlib.Path(config_path).read_text(encoding="utf-8")
    context: dict = {}
    exec(config_content, context)  # pylint: disable=exec-used

    config = context["config"]
    assert isinstance(config, Config)
    return config


def main():
    args = _parse_args(output_required=True)
    config = _load_config(args.config)
    output_json = generate_output(config)

    if args.output is not None:
        with open(args.output, "w", encoding="utf-8") as f:
            json.dump(output_json, f)


if __name__ == "__main__":
    main()



import pathlib
from typing import Dict

import pytest
from sphinx.testing.path import path as SphinxPath


pytest_plugins = ("sphinx.testing.fixtures",)


@pytest.fixture
def immaterial_make_app(tmp_path: pathlib.Path, make_app):

    conf = """
extensions = [
    "sphinx_immaterial",
]
html_theme = "sphinx_immaterial"
"""

    def make(files: Dict[str, str], extra_conf: str = "", **kwargs):
        (tmp_path / "conf.py").write_text(conf + extra_conf, encoding="utf-8")
        for filename, content in files.items():
            (tmp_path / filename).write_text(content, encoding="utf-8")
        app = make_app(srcdir=SphinxPath(str(tmp_path)), **kwargs)
        app.pdb = True
        return app

    yield make



import asyncio
import concurrent.futures
import hashlib
import json
import os
import re
import tempfile
from typing import Dict, Optional, List, Set, Tuple
import urllib.parse

import appdirs
import requests
import sphinx.application
import sphinx.config
import sphinx.util.logging

logger = sphinx.util.logging.getLogger(__name__)


def get_url(
    cache_dir: str, url: str, headers: Optional[Dict[str, str]] = None
) -> bytes:
    if headers is None:
        headers = {}
    req_json = {"url": url, "headers": headers}
    req_json_encoded = json.dumps(req_json).encode("utf-8")
    req_key = hashlib.sha256(req_json_encoded).hexdigest()

    resp_path = os.path.join(cache_dir, f"{req_key}.response")
    try:
        with open(resp_path, "rb") as f:
            return f.read()
    except FileNotFoundError:
        pass

    logger.info("Fetching: %s with %r", url, headers)
    r = requests.get(  # pylint: disable=missing-timeout
        url, headers=headers, stream=True
    )
    r.raise_for_status()

    response_content = r.content

    # Write request.
    req_path = os.path.join(cache_dir, f"{req_key}.request")
    os.makedirs(cache_dir, exist_ok=True)
    with open(req_path, "wb") as f:
        f.write(req_json_encoded)

    # Write response
    temp_name = None

    try:
        with tempfile.NamedTemporaryFile(
            dir=cache_dir, suffix=".tmp", prefix=req_key + ".request.", delete=False
        ) as f:
            temp_name = f.name
            f.write(response_content)
        os.replace(temp_name, resp_path)
        temp_name = None

        return response_content
    finally:
        if temp_name is not None:
            try:
                os.remove(temp_name)
            except:  # pylint: disable=bare-except
                pass


def _get_default_cache_dir(config: sphinx.config.Config):
    cache_dir = os.environ.get("SPHINX_IMMATERIAL_EXTERNAL_RESOURCE_CACHE_DIR")
    if cache_dir is not None:
        return cache_dir

    return os.path.join(
        appdirs.user_cache_dir("sphinx_immaterial", "jbms"), "external_resources"
    )


_RESOURCE_CONFIG_KEY = "sphinx_immaterial_external_resource_cache_dir"

# https://stackoverflow.com/questions/25011533/google-font-api-uses-browser-detection-how-to-get-all-font-variations-for-font
_FONT_FORMAT_USER_AGENT = {
    "ttf": "Safari 3.1 Mozilla/5.0 (Macintosh; U; PPC Mac OS X 10_5_2; en-gb) AppleWebKit/526+ (KHTML, like Gecko) Version/3.1 iPhone",
    "woff": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36",
    # "Safari 6.0 Mozilla/5.0 (iPad; CPU OS 6_0 like Mac OS X) AppleWebKit/536.26 (KHTML, like Gecko) Version/6.0 Mobile/10A5355d Safari/8536.25",
    # 'woff2': 'Firefox 36.0 Mozilla/5.0 (Windows NT 6.3; rv:36.0) Gecko/20100101 Firefox/36.0',
}


_CSS_URL_PATTERN = re.compile(r"url\(([^\)]+)\)")

_FILE_EXT_PATTERN = re.compile(r".*(\.[^\.]+)")


def _extract_urls(css_content: bytes) -> Set[str]:
    urls = set()
    for m in _CSS_URL_PATTERN.finditer(css_content.decode("utf-8")):
        urls.add(m.group(1))
    return urls


def _adjust_css_urls(css_content: bytes, renamed_fonts: Dict[str, str]) -> str:
    css_text = css_content.decode("utf-8")
    return _CSS_URL_PATTERN.sub(
        lambda m: f"url(fonts/{renamed_fonts[m.group(1)]})", css_text
    )


_MAX_CONCURRENT_FETCHES = 128

_TTF_FONT_PATHS_KEY = "sphinx_immaterial_ttf_font_paths"


def add_google_fonts(app: sphinx.application.Sphinx, fonts: List[str]):

    cache_dir = getattr(app.config, _RESOURCE_CONFIG_KEY)
    static_dir = os.path.join(app.outdir, "_static")
    # _static path
    font_dir = os.path.join(static_dir, "fonts")
    os.makedirs(font_dir, exist_ok=True)

    with concurrent.futures.ThreadPoolExecutor(max_workers=32) as executor:

        def to_thread(fn, *args, **kwargs) -> asyncio.Future:
            return asyncio.wrap_future(executor.submit(fn, *args, **kwargs))

        async def fetch_font(font: str, style: str):
            css_url = f"https://fonts.googleapis.com/css?family={urllib.parse.quote(font)}:{style}"

            css_content_future = asyncio.gather(
                *[
                    to_thread(
                        get_url,
                        cache_dir,
                        css_url,
                        headers={"user-agent": user_agent},
                    )
                    for user_agent in _FONT_FORMAT_USER_AGENT.values()
                ]
            )

            css_content = dict(
                zip(_FONT_FORMAT_USER_AGENT.keys(), await css_content_future)
            )

            font_files = set()
            ttf_font_files = None

            for font_format, css_data in css_content.items():
                urls = _extract_urls(css_data)
                font_files.update(urls)
                if font_format == "ttf":
                    ttf_font_files = urls

            font_data_futures = asyncio.gather(
                *[to_thread(get_url, cache_dir, font_url) for font_url in font_files]
            )

            all_font_data = dict(zip(font_files, await font_data_futures))

            renamed_fonts = {}

            for font_url, font_data in all_font_data.items():
                h = hashlib.sha256(font_data).hexdigest()[:32]
                m = _FILE_EXT_PATTERN.fullmatch(font_url)
                if m is not None:
                    file_ext = m.group(1)
                else:
                    file_ext = ""
                new_name = h + file_ext
                renamed_fonts[font_url] = new_name
                with open(os.path.join(font_dir, new_name), "wb") as f:
                    f.write(font_data)
            adjusted_css_content = {
                font_format: _adjust_css_urls(css_data, renamed_fonts)
                for font_format, css_data in css_content.items()
            }

            ttf_font_path = None

            if ttf_font_files is None or len(ttf_font_files) != 1:
                logger.error(
                    "Expected 1 TTF font for %s:%s but received: %r",
                    font,
                    style,
                    css_content["ttf"],
                )
            else:
                ttf_font_url = next(iter(ttf_font_files))
                ttf_font_path = renamed_fonts[ttf_font_url]

            return adjusted_css_content, ttf_font_path

        async def do_fetch():
            css_future_keys = []
            css_futures = []
            for font in fonts:
                for style in ["300", "300i", "400", "400i", "700", "700i"]:
                    css_future_keys.append((font, style))
                    css_futures.append(fetch_font(font, style))
            css_content = dict(zip(css_future_keys, await asyncio.gather(*css_futures)))
            return css_content

        css_content = asyncio.run(do_fetch())

    # Write fonts css file
    ttf_font_paths = {}
    with open(os.path.join(static_dir, "google_fonts.css"), "w", encoding="utf-8") as f:
        for key, (css_format_content, ttf_font_path) in css_content.items():
            ttf_font_paths[key] = os.path.join(font_dir, ttf_font_path)
            for content in css_format_content.values():
                f.write("".join(re.split(r"(?:\s*\n\s*|/\*.*\*/)", content)))

    app.add_css_file("google_fonts.css")
    setattr(app, _TTF_FONT_PATHS_KEY, ttf_font_paths)


def get_ttf_font_paths(app: sphinx.application.Sphinx) -> Dict[Tuple[str, str], str]:
    return getattr(app, _TTF_FONT_PATHS_KEY)


def _builder_inited(app: sphinx.application.Sphinx):
    font_options = app.config["html_theme_options"]["font"]
    if not font_options:
        return
    add_google_fonts(app, list(font_options.values()))


def setup(app: sphinx.application.Sphinx):
    app.add_config_value(
        _RESOURCE_CONFIG_KEY,
        default=_get_default_cache_dir,
        rebuild="env",
        types=(str,),
    )

    app.connect("builder-inited", _builder_inited)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import re
import pytest
from sphinx_immaterial.apidoc.cpp import api_parser


def test_basic():

    config = api_parser.Config(
        input_path="a.cpp",
        input_content=b"""

/// This is the doc.
int foo(bool x, int y);

""",
    )

    output = api_parser.generate_output(config)
    entities = list(output["entities"].values())
    assert len(entities) == 1


def test_enable_if_transform():
    input_path = "a.cpp"

    input_content = rb"""
namespace std {
template <bool Condition, typename T = void>
using enable_if_t = T;
template <typename T, typename U>
constexpr inline bool is_convertible_v = false;
}

/// This is the doc.
template <typename U, typename T, typename = std::enable_if_t<std::is_convertible_v<U(*)[], T(*)[]>>>
int foo(T x);

"""

    config = api_parser.Config(
        input_path=input_path,
        allow_paths=[re.escape(input_path)],
        disallow_namespaces=[re.compile("^std$")],
        compiler_flags=["-std=c++17"],
        input_content=input_content,
    )

    output = api_parser.generate_output(config)

    entities = list(output["entities"].values())
    assert len(entities) == 1
    requires = entities[0].get("requires")
    assert requires


class TestCommentStrip:
    config = api_parser.Config(
        input_path="a.cpp",
    )

    def assert_output(self, expected: str):
        output = api_parser.generate_output(self.config)
        doc_strings = [v["doc"]["text"] for v in output.get("entities", {}).values()]
        assert expected in doc_strings

    @pytest.mark.parametrize(
        argnames="doc_str,expected",
        argvalues=[
            (b"//! This is a docstring.\nint var = 0;", "This is a docstring."),
            (b"/// This is a docstring.\nint var = 0;", "This is a docstring."),
            (b"/** This is a docstring. */\nint var = 0;", "This is a docstring."),
            (
                b"/*! This is a docstring. */\nint var = 0;",
                "This is a docstring.",
            ),
            (
                b"int var = 0; ///< This is a docstring.",
                "This is a docstring.",
            ),
            (
                b"/**\n * This is a docstring.\n */\nint var = 0;",
                "\nThis is a docstring.",
            ),
            (
                b"/*!\n * This is a docstring.\n */\nint var = 0;",
                "\nThis is a docstring.",
            ),
            (
                b"// Skip this.\n/// This is a docstring.\n// Skip this.\nint var = 0;",
                "This is a docstring.",
            ),
            (
                b"/* Skip\n * this.*/\n/**This is a docstring.*/\n"
                b"/*Skip this.*/\nint var = 0;",
                "This is a docstring.",
            ),
        ],
        ids=[
            "///",
            "//!",
            "/**",
            "/*!",
            "///<",
            "/** \n * \n */",
            "/*! \n * \n */",
            "// \n /// \n //",
            "/* \n * \n */\n /** */\n /* */",
        ],
    )
    def test_comment_styles(self, doc_str: bytes, expected: str):
        self.config.input_content = doc_str
        self.assert_output(expected)


def test_function_fields():
    config = api_parser.Config(
        input_path="a.cpp",
        input_content="""
/// @brief A dummy function for tests.
///
/// \\details A more detailed paragraph.
///
/// @param arg1 An arg passed by value.
/// \\param[in] arg2 An unaltered arg passed by reference.
/// @param[in, out] arg3 An arg passed by reference that gets altered.
/// @retval NULL If unsuccessful.
/// @returns A flag indicating success.
/// \\tparam T A template parameter.
template<typename T>
int function(T arg1, T &arg2, T &arg3);
""",
    )

    output = api_parser.generate_output(config)
    entities = output.get("entities", {})
    doc_str = ""
    for entity in entities.values():  # type: dict
        if entity.get("doc"):
            doc_str = entity["doc"]["text"]
            break
    else:
        raise AttributeError("no doc string found.")
    expected = "\n".join(
        [
            "A dummy function for tests.",
            "",
            "A more detailed paragraph.",
            "",
            ":param arg1: An arg passed by value.",
            ":param arg2[in]: An unaltered arg passed by reference.",
            ":param arg3[in, out]: An arg passed by reference that gets altered.",
            ":retval NULL: If unsuccessful.",
            ":returns: A flag indicating success.",
            ":tparam T: A template parameter.",
        ]
    )
    assert expected == doc_str
